﻿<###################################################
 #                                                 #
 #  Copyright (c) Microsoft. All rights reserved.  #
 #                                                 #
 ##################################################>

<#
.Synopsis
   Initializes the module with the necessary information to call Graph APIs in a user context.
#>
function Initialize-GraphEnvironment
{
    [CmdletBinding(DefaultParameterSetName='Credential_AAD')]
    param
    (
        # The directory tenant identifier of the primary issuer in which Graph API calls should be made.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $DirectoryTenantId,

        # The user credential with which to acquire an access token targeting Graph.
        [Parameter(ParameterSetName='Credential_AAD')]
        [Parameter(ParameterSetName='Credential_ADFS')]
        [ValidateNotNull()]
        [pscredential] $UserCredential = $null,

        # Indicates that the script should prompt the user to input a credential with which to acquire an access token targeting Graph.
        [Parameter(ParameterSetName='Credential_AAD')]
        [Parameter(ParameterSetName='Credential_ADFS')]
        [switch] $PromptForUserCredential,

        # The refresh token to use to acquire an access token targeting Graph.
        [Parameter(ParameterSetName='RefreshToken_AAD')]
        [Parameter(ParameterSetName='RefreshToken_ADFS')]
        [ValidateNotNull()]
        [SecureString] $RefreshToken = $null,

        # The client identifier (application identifier) of a service principal with which to acquire an access token targeting Graph.
        [Parameter(ParameterSetName='ServicePrincipal_AAD')]
        [ValidateNotNullOrEmpty()]
        [string] $ClientId = $null,

        # The client certificate of a service principal with which to acquire an access token targeting Graph.
        [Parameter(ParameterSetName='ServicePrincipal_AAD')]
        [ValidateNotNull()]
        [System.Security.Cryptography.X509Certificates.X509Certificate2] $ClientCertificate = $null,

        # The name of the supported Cloud Environment in which the target Graph Service is available.
        [Parameter(ParameterSetName='Credential_AAD')]
        [Parameter(ParameterSetName='RefreshToken_AAD')]
        [Parameter(ParameterSetName='ServicePrincipal_AAD')]
        [ValidateSet('AzureCloud', 'AzureChinaCloud', 'AzureUSGovernment', 'AzureGermanCloud')]
        [string] $Environment = 'AzureCloud',

        # The fully-qualified domain name of the ADFS service (e.g. "adfs.azurestack.local").
        [Parameter(Mandatory=$true, ParameterSetName='Credential_ADFS')]
        [Parameter(Mandatory=$true, ParameterSetName='RefreshToken_ADFS')]
        [ValidateNotNullOrEmpty()]
        [string] $AdfsFqdn,

        # The fully-qualified domain name of the on-premise Graph service (e.g. "graph.azurestack.local").
        [Parameter(Mandatory=$true, ParameterSetName='Credential_ADFS')]
        [Parameter(Mandatory=$true, ParameterSetName='RefreshToken_ADFS')]
        [ValidateNotNullOrEmpty()]
        [string] $GraphFqdn
    )

    if ($AdfsFqdn)
    {
        $Environment = 'ADFS'
        Write-Warning "Parameters for ADFS have been specified; please note that only a subset of Graph APIs are available to be used in conjuction with ADFS."
    }

    if ($PromptForUserCredential)
    {
        $UserCredential = Get-Credential -Message "Please provide a credential used to access Graph. Must support non-interactive authentication flows."
    }

    if ($UserCredential)
    {
        Write-Verbose "Initializing the module to use Graph environment '$Environment' for user '$($UserCredential.UserName)' in directory tenant '$DirectoryTenantId'." -Verbose
    }
    elseif ($RefreshToken)
    {
        Write-Verbose "Initializing the module to use Graph environment '$Environment' (with refresh token) in directory tenant '$DirectoryTenantId'." -Verbose
    }
    elseif ($ClientId -and $ClientCertificate)
    {
        Write-Verbose "Initializing the module to use Graph environment '$Environment' for service principal '$($ClientId)' in directory tenant '$DirectoryTenantId' with certificate $($ClientCertificate.Thumbprint)." -Verbose
    }
    else
    {
        Write-Warning "A user credential, refresh token, or service principal info was not provided. Graph API calls cannot be made until one is provided. Please run 'Initialize-GraphEnvironment' again with valid credentials."
    }

    $graphEnvironmentTemplate = @{}
    $graphEnvironmentTemplate += switch ($Environment)
    {
        'AzureCloud'
        {
            @{
                GraphVersion  = "1.6"
                GraphResource = "https://graph.windows.net/"

                IssuerTemplate = "https://sts.windows.net/{0}/"

                LoginEndpoint = [Uri]"https://login.microsoftonline.com/$DirectoryTenantId"
                GraphEndpoint = [Uri]"https://graph.windows.net/$DirectoryTenantId"

                LoginBaseEndpoint = [Uri]"https://login.microsoftonline.com/"
                GraphBaseEndpoint = [Uri]"https://graph.windows.net/"

                FederationMetadataEndpoint = [Uri]"https://login.microsoftonline.com/$DirectoryTenantId/federationmetadata/2007-06/federationmetadata.xml"
                OpenIdMetadata             = [Uri]"https://login.microsoftonline.com/$DirectoryTenantId/.well-known/openid-configuration"

                AadPermissions = [HashTable]@{
                    AccessDirectoryAsSignedInUser      = "a42657d6-7f20-40e3-b6f0-cee03008a62a"
                    EnableSignOnAndReadUserProfiles    = "311a71cc-e848-46a1-bdf8-97ff7156d8e6"
                    ReadAllGroups                      = "6234d376-f627-4f0f-90e0-dff25c5211a3"
                    ReadAllUsersBasicProfile           = "cba73afc-7f69-4d86-8450-4978e04ecd1a"
                    ReadAllUsersFullProfile            = "c582532d-9d9e-43bd-a97c-2667a28ce295"
                    ReadDirectoryData                  = "5778995a-e1bf-45b8-affa-663a9f3f4d04"
                    ManageAppsThatThisAppCreatesOrOwns = "824c81eb-e3f8-4ee6-8f6d-de7f50d565b7"
                }
            }
        }

        'AzureChinaCloud'
        {
            @{
                GraphVersion  = "1.6"
                GraphResource = "https://graph.chinacloudapi.cn/"

                IssuerTemplate = "https://sts.chinacloudapi.cn/{0}/"

                LoginEndpoint = [Uri]"https://login.chinacloudapi.cn/$DirectoryTenantId"
                GraphEndpoint = [Uri]"https://graph.chinacloudapi.cn/$DirectoryTenantId"

                LoginBaseEndpoint = [Uri]"https://login.chinacloudapi.cn/"
                GraphBaseEndpoint = [Uri]"https://graph.chinacloudapi.cn/"

                FederationMetadataEndpoint = [Uri]"https://login.chinacloudapi.cn/$DirectoryTenantId/federationmetadata/2007-06/federationmetadata.xml"
                OpenIdMetadata             = [Uri]"https://login.chinacloudapi.cn/$DirectoryTenantId/.well-known/openid-configuration"

                AadPermissions = [HashTable]@{
                    AccessDirectoryAsSignedInUser      = "a42657d6-7f20-40e3-b6f0-cee03008a62a"
                    EnableSignOnAndReadUserProfiles    = "311a71cc-e848-46a1-bdf8-97ff7156d8e6"
                    ReadAllGroups                      = "6234d376-f627-4f0f-90e0-dff25c5211a3"
                    ReadAllUsersBasicProfile           = "cba73afc-7f69-4d86-8450-4978e04ecd1a"
                    ReadAllUsersFullProfile            = "c582532d-9d9e-43bd-a97c-2667a28ce295"
                    ReadDirectoryData                  = "5778995a-e1bf-45b8-affa-663a9f3f4d04"
                    ManageAppsThatThisAppCreatesOrOwns = "b55274d3-3582-44e3-83ae-ed7873d1111d" # This permission is different than in 'AzureCloud'
                }
            }
        }

        'AzureUSGovernment'
        {
            @{
                GraphVersion  = "1.6"
                GraphResource = "https://graph.windows.net/"

                IssuerTemplate = "https://sts.windows.net/{0}/"

                LoginEndpoint = [Uri]"https://login-us.microsoftonline.com/$DirectoryTenantId"
                GraphEndpoint = [Uri]"https://graph.windows.net/$DirectoryTenantId"

                LoginBaseEndpoint = [Uri]"https://login-us.microsoftonline.com/"
                GraphBaseEndpoint = [Uri]"https://graph.windows.net/"

                FederationMetadataEndpoint = [Uri]"https://login-us.microsoftonline.com/$DirectoryTenantId/federationmetadata/2007-06/federationmetadata.xml"
                OpenIdMetadata             = [Uri]"https://login-us.microsoftonline.com/$DirectoryTenantId/.well-known/openid-configuration"

                AadPermissions = [HashTable]@{
                    AccessDirectoryAsSignedInUser      = "a42657d6-7f20-40e3-b6f0-cee03008a62a"
                    EnableSignOnAndReadUserProfiles    = "311a71cc-e848-46a1-bdf8-97ff7156d8e6"
                    ReadAllGroups                      = "6234d376-f627-4f0f-90e0-dff25c5211a3"
                    ReadAllUsersBasicProfile           = "cba73afc-7f69-4d86-8450-4978e04ecd1a"
                    ReadAllUsersFullProfile            = "c582532d-9d9e-43bd-a97c-2667a28ce295"
                    ReadDirectoryData                  = "5778995a-e1bf-45b8-affa-663a9f3f4d04"
                    ManageAppsThatThisAppCreatesOrOwns = "824c81eb-e3f8-4ee6-8f6d-de7f50d565b7"
                }
            }
        }

        'AzureGermanCloud'
        {
            @{
                GraphVersion  = "1.6"
                GraphResource = "https://graph.cloudapi.de/"

                IssuerTemplate = "https://sts.microsoftonline.de/{0}/"

                LoginEndpoint = [Uri]"https://login.microsoftonline.de/$DirectoryTenantId"
                GraphEndpoint = [Uri]"https://graph.cloudapi.de/$DirectoryTenantId"

                LoginBaseEndpoint = [Uri]"https://login.microsoftonline.de/"
                GraphBaseEndpoint = [Uri]"https://graph.cloudapi.de/"

                FederationMetadataEndpoint = [Uri]"https://login.microsoftonline.de/$DirectoryTenantId/federationmetadata/2007-06/federationmetadata.xml"
                OpenIdMetadata             = [Uri]"https://login.microsoftonline.de/$DirectoryTenantId/.well-known/openid-configuration"

                AadPermissions = [HashTable]@{
                    AccessDirectoryAsSignedInUser      = "a42657d6-7f20-40e3-b6f0-cee03008a62a"
                    EnableSignOnAndReadUserProfiles    = "311a71cc-e848-46a1-bdf8-97ff7156d8e6"
                    ReadAllGroups                      = "6234d376-f627-4f0f-90e0-dff25c5211a3"
                    ReadAllUsersBasicProfile           = "cba73afc-7f69-4d86-8450-4978e04ecd1a"
                    ReadAllUsersFullProfile            = "c582532d-9d9e-43bd-a97c-2667a28ce295"
                    ReadDirectoryData                  = "5778995a-e1bf-45b8-affa-663a9f3f4d04"
                    ManageAppsThatThisAppCreatesOrOwns = "824c81eb-e3f8-4ee6-8f6d-de7f50d565b7"
                }
            }
        }

        'ADFS'
        {
            @{
                GraphVersion  = "2016-01-01"
                GraphResource = "https://$GraphFqdn/"

                IssuerTemplate = "https://$AdfsFqdn/adfs/{0}/"

                LoginEndpoint = [Uri]"https://$AdfsFqdn/adfs/$DirectoryTenantId"
                GraphEndpoint = [Uri]"https://$GraphFqdn/$DirectoryTenantId"

                LoginBaseEndpoint = [Uri]"https://$AdfsFqdn/adfs/"
                GraphBaseEndpoint = [Uri]"https://$GraphFqdn/"

                FederationMetadataEndpoint = [Uri]"https://$AdfsFqdn/federationmetadata/2007-06/federationmetadata.xml"
                OpenIdMetadata             = [Uri]"https://$AdfsFqdn/adfs/$DirectoryTenantId/.well-known/openid-configuration"
            }
        }

        default
        {
            throw New-Object NotImplementedException("Unknown environment type '$Environment'")
        }
    }

    # Note: if this data varies from environment to environment, declare it in switch above
    $graphEnvironmentTemplate += @{
        Environment       = $Environment
        DirectoryTenantId = $DirectoryTenantId

        User = [pscustomobject]@{
            Credential            = $UserCredential
            DirectoryTenantId     = $DirectoryTenantId
            AccessToken           = $null
            RefreshToken          = $RefreshToken
            AccessTokenUpdateTime = $null
            AccessTokenExpiresIn  = $null
            ClientRequestId       = [guid]::NewGuid().ToString()
            ServicePrincipal = [pscustomobject]@{
                ClientId    = $ClientId
                Certificate = $ClientCertificate
            }
        }

        Applications = [pscustomobject]@{
            LegacyPowerShell            = [pscustomobject]@{ Id = "0a7bdc5c-7b57-40be-9939-d4c5fc7cd417" }
            PowerShell                  = [pscustomobject]@{ Id = "1950a258-227b-4e31-a9cf-717495945fc2" }
            WindowsAzureActiveDirectory = [pscustomobject]@{ Id = "00000002-0000-0000-c000-000000000000" }
            VisualStudio                = [pscustomobject]@{ Id = "872cd9fa-d31f-45e0-9eab-6e460a02d1f1" }
            AzureCLI                    = [pscustomobject]@{ Id = "04b07795-8ddb-461a-bbee-02f9e1bf7b46" }
        }

        AadPermissionScopes = [HashTable]@{
            AccessDirectoryAsSignedInUser      = "Directory.AccessAsUser.All"
            EnableSignOnAndReadUserProfiles    = "User.Read"
            ReadAllGroups                      = "Group.Read.All"
            ReadAllUsersBasicProfile           = "User.ReadBasic.All"
            ReadAllUsersFullProfile            = "User.Read.All"
            ReadDirectoryData                  = "Directory.Read.All"
            ManageAppsThatThisAppCreatesOrOwns = "Application.ReadWrite.OwnedBy"
        }
    }

    if ($AdfsFqdn)
    {
        $graphEnvironmentTemplate.Applications = [pscustomobject]@{}
    }

    $Script:GraphEnvironment = [pscustomobject]$graphEnvironmentTemplate
    Write-Verbose "Graph Environment initialized: client-request-id: $($Script:GraphEnvironment.User.ClientRequestId)" -Verbose

    # Attempt to log-in the user
    if ($UserCredential -or $RefreshToken -or ($ClientId -and $ClientCertificate))
    {
        Update-GraphAccessToken -Verbose
    }
}

<#
.Synopsis
   Gets the Graph environment information, including endpoints and Graph API version.
#>
function Get-GraphEnvironmentInfo
{
    [CmdletBinding()]
    [OutputType([pscustomobject])]
    param
    (
    )

    # Return a cloned copy of the environment data without the user information
    $Script:GraphEnvironment | Select -Property * -ExcludeProperty User | ConvertTo-Json -Depth 10 | ConvertFrom-Json | Write-Output
}

<#
.Synopsis
   Asserts that Initialize-GraphEnvironment has been called in the current runspace.
#>
function Assert-GraphEnvironmentIsInitialized
{
    if (-not $Script:GraphEnvironment -or
        -not ($Script:GraphEnvironment.User.Credential -or
             $Script:GraphEnvironment.User.RefreshToken -or
             ($Script:GraphEnvironment.User.ServicePrincipal.ClientId -and $Script:GraphEnvironment.User.ServicePrincipal.Certificate)))
    {
        throw New-Object InvalidOperationException("The graph environment has not yet been initialized. Please run 'Initialize-GraphEnvironment' with a valid credential or refresh token.")
    }
}

<#
.Synopsis
   Asserts that a connection can be established to the initialized graph environment.
#>
function Assert-GraphConnection
{
    [CmdletBinding()]
    param
    (
    )

    Assert-GraphEnvironmentIsInitialized

    Write-Verbose "Testing connection to graph environment using endpoint '$($Script:GraphEnvironment.OpenIdMetadata)'" -Verbose

    try
    {
        $response      = Invoke-WebRequest -UseBasicParsing -Uri $Script:GraphEnvironment.OpenIdMetadata -Verbose -TimeoutSec 90 -ErrorAction Stop
        $traceResponse = $response | Select StatusCode,StatusDescription,@{n='Content';e={ConvertFrom-Json $_.Content}} | ConvertTo-Json
        Write-Verbose "Verified a successful connection to the graph service; response received: $traceResponse" -Verbose
    }
    catch
    {
        # In the case of errors, there is no response returned to caller (even when error action is set to ignore, continue, etc.) so we extract the response from the thrown exception (if there is one)
        $traceResponse = $_.Exception.Response | Select Method,ResponseUri,StatusCode,StatusDescription,IsFromCache,LastModified | ConvertTo-Json

        if ($_.Exception.Response.StatusCode -and $_.Exception.Response.StatusCode -lt 500)
        {
            # This means we received a valid response from graph (our connection is good) but there was some other error in the call...
            Write-Warning "An unexpected error response was received while validating a connection to the graph service: $_`r`n`r`nAdditional details: $traceResponse"
        }
        else
        {
            # Trace the message to verbose stream as well in case error is not traced in same file as other verbose logs
            $traceMessage = "An error occurred while trying to verify connection to the graph endpoint '$($Script:GraphEnvironment.OpenIdMetadata)': $_`r`n`r`nAdditional details: $traceResponse"
            Write-Verbose "ERROR: $traceMessage"

            throw New-Object System.InvalidOperationException($traceMessage)
        }
    }
}

<#
.Synopsis
   Acquires a token from Graph for the specified resource using the specified or initialized credential or refresh token.
#>
function Get-GraphToken
{
    [CmdletBinding(DefaultParametersetName='Credential')]
    [OutputType([pscustomobject])]
    param
    (
        # The resource for which to acquire a token.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [string] $Resource = $Script:GraphEnvironment.GraphResource,

        # The user credential with which to acquire an access token targeting the specified resource.
        [Parameter(Mandatory=$true, ParameterSetName='Credential')]
        [ValidateNotNull()]
        [pscredential] $Credential = $null,

        # The refresh token to use to acquire an access token targeting the specified resource.
        [Parameter(Mandatory=$true, ParameterSetName='RefreshToken')]
        [ValidateNotNullOrEmpty()]
        [SecureString] $RefreshToken = $null,

        # The client identifier (application identifier) of a service principal with which to acquire an access token targeting the specified resource.
        [Parameter(Mandatory=$true, ParameterSetName='ServicePrincipal')]
        [ValidateNotNull()]
        [string] $ClientId = $null,

        # The client certificate of a service principal with which to acquire an access token targeting targeting the specified resource.
        [Parameter(Mandatory=$true, ParameterSetName='ServicePrincipal')]
        [ValidateNotNull()]
        [System.Security.Cryptography.X509Certificates.X509Certificate2] $ClientCertificate = $null,

        # Indicates whether the user credential or refresh token should be used from the initialized environment data.
        [Parameter(Mandatory=$true, ParameterSetName='Environment')]
        [Switch] $UseEnvironmentData
    )

    Assert-GraphConnection

    if ($UseEnvironmentData)
    {
        if ($Script:GraphEnvironment.User.Credential)
        {
            $Credential = $Script:GraphEnvironment.User.Credential
        }
        elseif ($Script:GraphEnvironment.User.RefreshToken)
        {
            $RefreshToken = $Script:GraphEnvironment.User.RefreshToken
        }
        elseif ($Script:GraphEnvironment.User.ServicePrincipal.ClientId -and $Script:GraphEnvironment.User.ServicePrincipal.Certificate)
        {
            $ClientId          = $Script:GraphEnvironment.User.ServicePrincipal.ClientId
            $ClientCertificate = $Script:GraphEnvironment.User.ServicePrincipal.Certificate
        }
    }

    $requestBody = @{ resource = $Resource }

    if ($Credential)
    {
        $requestBody += @{
            client_id  = $Script:GraphEnvironment.Applications.PowerShell.Id
            grant_type = 'password'
            scope      = 'openid'
            username   = $Credential.UserName
            password   = $Credential.GetNetworkCredential().Password
        }

        Write-Verbose "Attempting to acquire a token for resource '$Resource' using a user credential '$($Credential.UserName)'"
    }
    elseif ($RefreshToken)
    {
        $requestBody += @{
            client_id     = $Script:GraphEnvironment.Applications.PowerShell.Id
            grant_type    = 'refresh_token'
            scope         = 'openid'
            refresh_token = (New-Object System.Net.NetworkCredential('refreshToken', $RefreshToken)).Password
        }

        Write-Verbose "Attempting to acquire a token for resource '$Resource' using a refresh token"
    }
    elseif ($ClientId -and $ClientCertificate)
    {
        $params = @{
            ClientCertificate = $ClientCertificate
            ClientId          = $ClientId
            Audience          = "$($Script:GraphEnvironment.LoginEndpoint)".Trim('/') + '/oauth2/token'
        }
        $jwt = New-SelfSignedJsonWebToken @params

        $requestBody += @{
            client_id             = $ClientId
            grant_type            = 'client_credentials'
            client_assertion_type = 'urn:ietf:params:oauth:client-assertion-type:jwt-bearer'
            client_assertion      = $jwt
        }

        Write-Verbose "Attempting to acquire a token for resource '$Resource' using a service principal credential (id='$($ClientId)', thumbprint='$($ClientCertificate.Thumbprint)')"
    }
    else
    {
        throw New-Object InvalidOperationException("A valid user credential or refresh token is required to acquire a token from Graph service. Please run 'Initialize-GraphEnvironment' with a valid user credential or refresh token, or try the command again with one of the necessary values.")
    }

    $loginUserRequest = @{
        Method       = [Microsoft.PowerShell.Commands.WebRequestMethod]::Post
        BaseEndpoint = $Script:GraphEnvironment.LoginEndpoint
        ApiPath      = 'oauth2/token'
        ContentType  = "application/x-www-form-urlencoded"
        Body         = ConvertTo-QueryString $requestBody
    }

    $response = Invoke-GraphApi @loginUserRequest -UpdateGraphAccessTokenIfNecessary:$false
    Write-Output $response
}

<#
.Synopsis
   Updates the user Graph access token using the configured Graph Environment details.
#>
function Update-GraphAccessToken
{
    [CmdletBinding()]
    param
    (
    )

    # Attempt to log-in the user
    $response = Get-GraphToken -UseEnvironmentData

    $Script:GraphEnvironment.User.AccessToken           = $response.access_token
    $Script:GraphEnvironment.User.RefreshToken          = if ($response.refresh_token) { ConvertTo-SecureString $response.refresh_token -AsPlainText -Force } else { $null }
    $Script:GraphEnvironment.User.AccessTokenUpdateTime = [DateTime]::UtcNow
    $Script:GraphEnvironment.User.AccessTokenExpiresIn  = $response.expires_in
}

<#
.Synopsis
   Makes an API call to the graph service.
#>
function Invoke-GraphApi
{
    [CmdletBinding()]
    param
    (
        # The request method.
        [Parameter()]
        [ValidateNotNull()]
        [Microsoft.PowerShell.Commands.WebRequestMethod] $Method = [Microsoft.PowerShell.Commands.WebRequestMethod]::Get,

        # The API path to call.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $ApiPath,

        # The (additional) query parameters to include in the request.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [HashTable] $QueryParameters = @{},

        # The custom (additional) headers to include in the request.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [HashTable] $CustomHeaders = @{},

        # The base endpoint of the Graph service to call.
        [Parameter()]
        [ValidateNotNull()]
        [Uri] $BaseEndpoint = $Script:GraphEnvironment.GraphEndpoint,

        # The content type of the request.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [string] $ContentType = "application/json",

        # The body of the request.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [string] $Body = $null,

        # The OCP Session Key used to route subsequent requests
        [Parameter()]
        [string] $SessionKey = $Global:GraphAPI_LastResponse.Headers.'ocp-aad-session-key',

        # Indicates whether the Graph access token should be automatically refreshed if it is close to or has already expired (true by default).
        [Switch] $UpdateGraphAccessTokenIfNecessary = $true,

        # Indicates whether responses containing "OData NextLinks" should be automatically called and aggregated (true by default).
        [Switch] $AggregateNextLinkData = $true
    )

    Assert-GraphEnvironmentIsInitialized

    if ($UpdateGraphAccessTokenIfNecessary)
    {
        $secondsSinceTokenWasLastUpdated = [DateTime]::UtcNow.Subtract($Script:GraphEnvironment.User.AccessTokenUpdateTime).TotalSeconds
        if ($secondsSinceTokenWasLastUpdated -gt ($Script:GraphEnvironment.User.AccessTokenExpiresIn - 90))
        {
            Write-Verbose "Updating graph access token"
            Update-GraphAccessToken
        }
    }

    # Initialize the request parameters
    $graphApiRequest = @{
        Method      = $Method
        Uri         = '{0}/{1}' -f $BaseEndpoint.AbsoluteUri.TrimEnd('/'), $ApiPath.Trim('/')
        Headers     = @{
            "User-Agent"        = "Microsoft AzureStack Graph PowerShell"
            "client-request-id" = $Script:GraphEnvironment.User.ClientRequestId
        }
        ContentType = $ContentType
    }

    # Set the authorization header if we already have an access token
    if ($Script:GraphEnvironment.User.AccessToken)
    {
        $graphApiRequest.Headers["Authorization"] = "Bearer $($Script:GraphEnvironment.User.AccessToken)"
    }

    # Add session key header if present to route subsequent requests to the same replica
    if ($SessionKey)
    {
        $graphApiRequest.Headers['ocp-aad-session-key'] = $SessionKey
    }

    # Apply any custom headers specified by the caller (overriding defaults)
    foreach ($header in $CustomHeaders.GetEnumerator())
    {
        $graphApiRequest.Headers[$header.Key] = $header.Value
    }

    # Initialize the query string parameters
    $queryParams = @{ 'api-version' = $Script:GraphEnvironment.GraphVersion }

    # Apply any custom query parameters specified by the caller (overriding defaults)
    foreach ($queryParam in $QueryParameters.GetEnumerator())
    {
        $queryParams[$queryParam.Key] = $queryParam.Value
    }

    $graphApiRequest.Uri += '?{0}' -f (ConvertTo-QueryString $queryParams)

    if ($Body)
    {
        $graphApiRequest['Body'] = $Body
    }

    # Make the API call, and auto-follow / aggregate next-link responses
    try
    {
        $Global:GraphAPI_LastResponse = Invoke-WebRequest @graphApiRequest -UseBasicParsing -TimeoutSec 90 -ErrorAction Stop
        $response = $Global:GraphAPI_LastResponse.Content | ConvertFrom-Json
    }
    catch
    {
        # In the case of errors, there is no response returned to caller (even when error action is set to ignore, continue, etc.) so we extract the response from the thrown exception (if there is one)
        $traceResponse = $_.Exception.Response | Select Method,ResponseUri,StatusCode,StatusDescription,IsFromCache,LastModified | ConvertTo-Json

        # Trace the message to verbose stream as well in case error is not traced in same file as other verbose logs
        $traceMessage = "An error occurred while trying to make a graph API call: $_`r`n`r`nAdditional details: $traceResponse"
        Write-Verbose "ERROR: $traceMessage"

        throw New-Object System.InvalidOperationException($traceMessage)
    }

    if ((-not $response."odata.nextLink") -or (-not $AggregateNextLinkData))
    {
        # Preserve most-recent OCP session key
        if (-not $Global:GraphAPI_LastResponse.Headers.'ocp-aad-session-key')
        {
            $Global:GraphAPI_LastResponse.Headers.'ocp-aad-session-key' = $SessionKey
        }

        Write-Output $response
    }
    else
    {
        $originalResponse = $response | Select -Property * -ExcludeProperty "odata.nextLink"

        while ($response."odata.nextLink")
        {
            # Delay briefly between nextlink calls as they can overwhelm the proxy and / or AAD...
            Start-Sleep -Milliseconds 100

            # Note: the next link URI cannot be used directly as it does not preserve all the query parameters (such as API version)

            # Initialize the query string parameters
            $queryParams = @{ 'api-version' = $Script:GraphEnvironment.GraphVersion }

            # Apply any custom query parameters specified by the caller (overriding defaults)
            foreach ($queryParam in $QueryParameters.GetEnumerator())
            {
                $queryParams[$queryParam.Key] = $queryParam.Value
            }

            # Apply the next link query params (overriding others as applicable)
            $nextLinkQueryParams = [regex]::Unescape($response."odata.nextLink".Split('?')[1])
            $query = [System.Web.HttpUtility]::ParseQueryString($nextLinkQueryParams)
            foreach ($key in $query.Keys)
            {
                $queryParams[$key] = $query[$key]
            }

            # Note: sometimes, the next link URL is relative, and other times it is absolute!
            $absoluteOrRelativeAddress = $response."odata.nextLink".Split('?')[0].TrimStart('/')

            $graphApiRequest.Uri = if ($absoluteOrRelativeAddress.StartsWith("https"))
            {
                '{0}?{1}' -f @($absoluteOrRelativeAddress, (ConvertTo-QueryString $queryParams))
            }
            else
            {
                '{0}/{1}?{2}' -f @(
                    $BaseEndpoint.AbsoluteUri.TrimEnd('/'),
                    $absoluteOrRelativeAddress,
                    (ConvertTo-QueryString $queryParams))
            }

            try
            {
                $Global:GraphAPI_LastResponse = Invoke-WebRequest @graphApiRequest -UseBasicParsing -TimeoutSec 90 -ErrorAction Stop
                $response = $Global:GraphAPI_LastResponse.Content | ConvertFrom-Json
                $originalResponse.Value += @($response.Value)
            }
            catch
            {
                # In the case of errors, there is no response returned to caller (even when error action is set to ignore, continue, etc.) so we extract the response from the thrown exception (if there is one)
                $traceResponse = $_.Exception.Response | Select Method,ResponseUri,StatusCode,StatusDescription,IsFromCache,LastModified | ConvertTo-Json

                # Trace the message to verbose stream as well in case error is not traced in same file as other verbose logs
                $traceMessage = "An error occurred while trying to make a graph API call: $_`r`n`r`nAdditional details: $traceResponse"
                Write-Verbose "ERROR: $traceMessage"

                throw New-Object System.InvalidOperationException($traceMessage)
            }
        }

        # Preserve most-recent OCP session key
        if (-not $Global:GraphAPI_LastResponse.Headers.'ocp-aad-session-key')
        {
            $Global:GraphAPI_LastResponse.Headers.'ocp-aad-session-key' = $SessionKey
        }

        Write-Output $originalResponse
    }
}

<#
.Synopsis
   Gets the verified domain of the currently-configured directory tenant.
#>
function Get-GraphDefaultVerifiedDomain
{
    [CmdletBinding()]
    [OutputType([string])]
    param
    (
    )

    $tenant = Invoke-GraphApi -ApiPath "tenantDetails" -ErrorAction Stop
    $verifiedDomain = $tenant.value.verifiedDomains | Where initial -EQ $true | Select -First 1
    Write-Output $verifiedDomain.name
}

<#
.Synopsis
   Attempts to find an existing Graph application object, or returns null if no such application can be found.
#>
function Find-GraphApplication
{
    [CmdletBinding(DefaultParameterSetName='ByUri')]
    [OutputType([pscustomobject])]
    param
    (
        # The application identifier URI.
        [Parameter(Mandatory=$true, ParameterSetName='ByUri')]
        [ValidateNotNullOrEmpty()]
        [string] $AppUri,

        # The application identifer.
        [Parameter(Mandatory=$true, ParameterSetName='ById')]
        [ValidateNotNullOrEmpty()]
        [string] $AppId,

        # The application display name.
        [Parameter(Mandatory=$true, ParameterSetName='ByDisplayName')]
        [ValidateNotNullOrEmpty()]
        [string] $DisplayName
    )

    $filter = if ($DisplayName) {"displayName eq '$DisplayName'"} elseif($AppUri) {"identifierUris/any(i:i eq '$AppUri')"} else {"appId eq '$AppId'"}
    $response = Invoke-GraphApi -ApiPath "applications" -QueryParameters @{ '$filter' = $filter } -ErrorAction Stop
    Write-Output $response.value
}

<#
.Synopsis
   Gets an existing Graph application object (returns an error if the application is not found).
#>
function Get-GraphApplication
{
    [CmdletBinding(DefaultParameterSetName='ByUri')]
    [OutputType([pscustomobject])]
    param
    (
        # The application identifier URI.
        [Parameter(Mandatory=$true, ParameterSetName='ByUri')]
        [ValidateNotNullOrEmpty()]
        [string] $AppUri,

        # The application identifer.
        [Parameter(Mandatory=$true, ParameterSetName='ById')]
        [ValidateNotNullOrEmpty()]
        [string] $AppId
    )

    $application = Find-GraphApplication @PSBoundParameters
    if (-not $application)
    {
        Write-Error "Application with identifier '${AppUri}${AppId}' not found"
    }
    else
    {
        Write-Output $application
    }
}

<#
.Synopsis
   Removes the specified object from the Graph directory.
#>
function Remove-GraphObject
{
    [CmdletBinding()]
    param
    (
        # The identifier of the object to remove.
        [Parameter(Mandatory=$true, ValueFromPipelineByPropertyName=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $objectId
    )
    process
    {
        $null = Invoke-GraphApi -Method Delete -ApiPath "directoryObjects/$objectId" -ErrorAction Stop
    }
}

<#
.Synopsis
   Attempts to find one or more existing Graph applications and service principals, by searching for service principals with a particular tag.
#>
function Find-GraphApplicationDataByServicePrincipalTag
{
    [CmdletBinding(DefaultParameterSetName='StartsWith')]
    param
    (
        # A string used to filter for application service principals that contain a tag that satifies the ODATA filter startswith '$StartsWith'.
        [Parameter(Mandatory=$true, ParameterSetName='StartsWith')]
        [string] $StartsWith = $null,

        # A string used to filter for application service principals that contain a tag that satifies the ODATA filter tag eq '$Equals'.
        [Parameter(Mandatory=$true, ParameterSetName='Equals')]
        [string] $Equals = $null,

        # Indicates whether application lookup should be skipped (and only to return the service principal data).
        [Parameter()]
        [Switch] $SkipApplicationLookup
    )

    $filter = if ($StartsWith) {"tags/any(tag:startswith(tag, '$StartsWith'))"} else {"tags/any(t:t eq '$Equals')"}

    $message = "Looking for service principals with filter '$filter'..."
    Write-Verbose $message
    Write-Progress -Activity $message
    $matchedServicePrincipals = (Invoke-GraphApi -ApiPath 'servicePrincipals()' -QueryParameters @{ '$filter' = $filter } -ErrorAction Stop).value
    Write-Verbose "Matched $(@($matchedServicePrincipals).Length) service principals using filter '$filter'"
    Write-Progress -Activity $message -Completed

    if ($SkipApplicationLookup)
    {
        @($matchedServicePrincipals) | ForEach { [pscustomobject]@{ ServicePrincipal = $_ } } | Write-Output
        return
    }

    $progress = 0
    $start    = Get-Date
    foreach ($matchedServicePrincipal in @($matchedServicePrincipals))
    {
        $progress++
        $elapsedSeconds  = ((Get-Date) - $start).TotalSeconds
        $progressRatio   = $progress / @($matchedServicePrincipals).Length
        $progressParams  = @{
            Activity         = "Looking-up AAD application objects"
            Status           = "Looking up application ($progress/$(@($matchedServicePrincipals).Length)) with appId='$($matchedServicePrincipal.appId)'"
            PercentComplete  = [Math]::Min(100, 100 * $progressRatio)
            SecondsRemaining = [Math]::Max(1, ($elapsedSeconds / $progressRatio) - $elapsedSeconds) # If it took 1 min for 10 items, and we have 100 items, it will likely take 10 minutes. Re-calculate this extrapolation on each iteration.
        }
        Write-Progress @progressParams

        $matchedApplication = Find-GraphApplication -AppId $matchedServicePrincipal.appId

        Write-Output ([pscustomobject]@{
            Application      = $matchedApplication
            ServicePrincipal = $matchedServicePrincipal
        })
    }

    Write-Progress -Activity "Looking-up AAD application objects" -Completed
}

<#
.Synopsis
   Attempts to retrieve a non-null graph object with a limited number of attempts.
#>
function Get-GraphObjectWithRetry
{
    [CmdletBinding()]
    [OutputType([pscustomobject])]
    param
    (
        # The script to run which should return an object. If the script throws an exception, the retry will NOT be performed. If the script returns any non-null value, it is considered successful, and the result will be returned.
        [Parameter(Mandatory=$true)]
        [ValidateNotNull()]
        [ScriptBlock] $GetScript,

        # The maximum number of attempts to make before return null.
        [Parameter(Mandatory=$true)]
        [ValidateRange(1, [int]::MaxValue)]
        [int] $MaxAttempts,

        # The delay in seconds between each subsequent attempt at running the script.
        [Parameter(Mandatory=$true)]
        [ValidateRange(1, [int]::MaxValue)]
        [int] $DelayInSecondsBetweenAttempts,

        # The minimal delay in seconds that the script should sleep before returning successfully (if a fatal error is encountered, delay is not enforced).
        [Parameter(Mandatory=$false)]
        [ValidateRange(1, [int]::MaxValue)]
        [int] $MinimumDelayInSeconds = 0
    )

    $result     = $null
    $attempts   = 0
    $totalDelay = 0

    do
    {
        $attempts++
        $message = if ($attempts -le 1) {'Attempting to retrieve graph object'} else {"[RETRY] Attempting to retrieve graph object (attempt $attempts of $MaxAttempts)"}
        Write-Verbose $message
        $result = & $GetScript

        if ((-not $result) -and ($attempts -lt $MaxAttempts))
        {
            Write-Verbose "[RETRY] Attempt $attempts failed, delaying for $DelayInSecondsBetweenAttempts seconds before retry..." -Verbose
            Start-Sleep -Seconds $DelayInSecondsBetweenAttempts
            $totalDelay += $DelayInSecondsBetweenAttempts
        }
    }
    while ((-not $result) -and ($attempts -lt $MaxAttempts))

    $remainingDelay = $MinimumDelayInSeconds - $totalDelay
    if ($remainingDelay -gt 0)
    {
        Write-Verbose "Delaying for an additional $remainingDelay seconds to ensure minimum delay of $MinimumDelayInSeconds seconds is achieved..." -Verbose
        Start-Sleep -Seconds $remainingDelay
    }

    Write-Output $result
}

<#
.Synopsis
   Gets an existing Graph application service principal object (returns an error if the application service principal object is not found).
#>
function Get-GraphApplicationServicePrincipal
{
    [CmdletBinding(DefaultParameterSetName='ByApplicationId')]
    [OutputType([pscustomobject])]
    param
    (
        # The application identifier.
        [Parameter(Mandatory=$true, ParameterSetName='ByApplicationId')]
        [ValidateNotNullOrEmpty()]
        [string] $ApplicationId,

        # The application identifier URI.
        [Parameter(Mandatory=$true, ParameterSetName='ByApplicationIdentifierUri')]
        [ValidateNotNullOrEmpty()]
        [string] $ApplicationIdentifierUri
    )

    $filter = if ($ApplicationId) { "appId eq '$ApplicationId'" } else { "servicePrincipalNames/any(c:c eq '$ApplicationIdentifierUri')" }

    $servicePrincipal = (Invoke-GraphApi -ApiPath servicePrincipals -QueryParameters @{ '$filter' = $filter }).value

    if (-not $servicePrincipal)
    {
        Write-Error "Application service principal with identifier '${ApplicationId}${ApplicationIdentifierUri}' not found"
    }
    else
    {
        Write-Output $servicePrincipal
    }
}

<#
.Synopsis
   Idempotently creates an application service principal in Graph with the specified properties.
#>
function Initialize-GraphApplicationServicePrincipal
{
    [CmdletBinding()]
    [OutputType([pscustomobject])]
    param
    (
        # The application identifier.
        [Parameter(Mandatory=$true, ValueFromPipelineByPropertyName=$true)]
        [ValidateNotNullOrEmpty()]
        [Alias('appId')]
        [string] $ApplicationId,

        # Optional: Tags to include in the application service principal.
        [Parameter()]
        [string[]] $Tags = @()
    )
    process
    {
        $getScript = { (Invoke-GraphApi -ApiPath servicePrincipals -QueryParameters @{ '$filter' = "appId eq '$ApplicationId'" }).value }

        # Create a service principal for the application (if one doesn't already exist)
        if (-not ($primaryServicePrincipal = & $getScript))
        {
            Write-Verbose "Creating service principal for application '$ApplicationId' in AAD..." -Verbose
            $servicePrincipalRequestBody = @{
                'odata.type'   = 'Microsoft.DirectoryServices.ServicePrincipal'
                accountEnabled = $true
                appId          = $ApplicationId
            }

            if ($Tags.Count -gt 0)
            {
                $servicePrincipalRequestBody += @{
                    'tags@odata.type' = 'Collection(Edm.String)'
                    tags = $Tags
                }
            }

            # Note: we poll for the object after creating it to avoid issues with replication delay
            $primaryServicePrincipal = Invoke-GraphApi -Method Post -ApiPath servicePrincipals -Body (ConvertTo-Json $servicePrincipalRequestBody)
            $primaryServicePrincipal = Get-GraphObjectWithRetry -GetScript $getScript -MaxAttempts 10 -DelayInSecondsBetweenAttempts 5 -MinimumDelayInSeconds 5
        }
        else
        {
            Write-Verbose "Service principal for application '$ApplicationId' already created in AAD directory tenant." -Verbose
            if ($Tags)
            {
                Update-GraphApplicationServicePrincipalTags -ApplicationId $ApplicationId -Tags $Tags
            }
        }

        Write-Output $primaryServicePrincipal
    }
}

<#
.Synopsis
   Updates the tags on an existing application service principal.
#>
function Update-GraphApplicationServicePrincipalTags
{
    [CmdletBinding(DefaultParameterSetName='ByApplicationId')]
    [OutputType([pscustomobject])]
    param
    (
        # The application identifier.
        [Parameter(Mandatory=$true, ParameterSetName='ByApplicationId')]
        [ValidateNotNullOrEmpty()]
        [string] $ApplicationId,

        # The application identifier URI.
        [Parameter(Mandatory=$true, ParameterSetName='ByApplicationIdentifierUri')]
        [ValidateNotNullOrEmpty()]
        [string] $ApplicationIdentifierUri,

        # Additional tags to include in the application service principal (if not already present).
        [Parameter(Mandatory=$true)]
        [string[]] $Tags = @()
    )

    $params = if ($ApplicationId) { @{ ApplicationId = $ApplicationId } } else { @{ ApplicationIdentifierUri = $ApplicationIdentifierUri } }
    $servicePrincipal = Get-GraphApplicationServicePrincipal @params

    $updatedTags = New-Object System.Collections.Generic.HashSet[string](,[string[]]$servicePrincipal.tags)
    foreach ($tag in $Tags)
    {
        if ($updatedTags.Add($tag))
        {
            Write-Verbose "Adding new tag to service principal: '$tag'"
        }
        else
        {
            Write-Verbose "Tag already present on service principal: '$tag'"
        }
    }

    Invoke-GraphApi -Method Patch -ApiPath "servicePrincipals/$($ServicePrincipal.objectId)" -Body (ConvertTo-Json ([pscustomobject]@{
        'tags@odata.type' = 'Collection(Edm.String)'
        tags = $updatedTags
    }))
}

<#
.Synopsis
   Idempotently creates an OAuth2Permission grant for an application service principal against another application service principal in Graph with the specified properties (service principals are created if they do not exist).
#>
function Initialize-GraphOAuth2PermissionGrant
{
    [CmdletBinding(DefaultParameterSetName='ClientAppId_ResourceAppId')]
    param
    (
        # The application identifier of the client application.
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_ResourceAppId')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_ResourceIdentifierUri')]
        [ValidateNotNullOrEmpty()]
        [string] $ClientApplicationId,

        # The application identifier URI of the client application.
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_ResourceAppId')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_ResourceIdentifierUri')]
        [ValidateNotNullOrEmpty()]
        [string] $ClientApplicationIdentifierUri,

        # The application identifier of the resource application.
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_ResourceAppId')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_ResourceAppId')]
        [ValidateNotNullOrEmpty()]
        [string] $ResourceApplicationId,

        # The application identifier URI of the resource application.
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_ResourceIdentifierUri')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_ResourceIdentifierUri')]
        [ValidateNotNullOrEmpty()]
        [string] $ResourceApplicationIdentifierUri,

        # The scope of the permission to grant (e.g. 'user_impersonation' [default value]).
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [string] $Scope = 'user_impersonation',

        # The consent type of the permission to grant (e.g. 'AllPrincipals' [default value]).
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [ValidateSet('AllPrincipals', 'Principal')]
        [string] $ConsentType = 'AllPrincipals'
    )

    # https://msdn.microsoft.com/en-us/library/azure/ad/graph/api/entity-and-complex-type-reference#oauth2permissiongrant-entity

    # Ensure the application service principals exist in the directory tenant

    $clientApplicationServicePrincipal = if ($ClientApplicationId)
    {
        Initialize-GraphApplicationServicePrincipal -ApplicationId $ClientApplicationId -ErrorAction Stop
    }
    else
    {
        Get-GraphApplicationServicePrincipal -ApplicationIdentifierUri $ClientApplicationIdentifierUri -ErrorAction Stop
    }

    $resourceApplicationServicePrincipal = if ($ResourceApplicationId)
    {
        Initialize-GraphApplicationServicePrincipal -ApplicationId $ResourceApplicationId -ErrorAction Stop
    }
    else
    {
        Get-GraphApplicationServicePrincipal -ApplicationIdentifierUri $ResourceApplicationIdentifierUri -ErrorAction Stop
    }

    # Note: value=Invalid characters found in scope. Allowed characters are %x20 / %x21 / %x23-5B / %x5D-7E
    $scopesToGrant = $Scope.Split(' ')

    # Note: the permission grants do not expire, but we must provide an expiration date to the API
    $queryParameters = @{
        '$filter' = "resourceId eq '$($resourceApplicationServicePrincipal.objectId)' and clientId eq '$($clientApplicationServicePrincipal.objectId)'"
        '$top'    = '500' # Note - there is an issue with this API if you use a large page size and the result set is large as well
    }
    $existingGrant = (Invoke-GraphApi -ApiPath oauth2PermissionGrants -QueryParameters $queryParameters).Value | Select -First 1

    if (-not $existingGrant)
    {
        Write-Verbose "Granting OAuth2Permission '$Scope' to application service principal '$($clientApplicationServicePrincipal.appDisplayName)' on behalf of application '$($resourceApplicationServicePrincipal.appDisplayName)'..." -Verbose
        $response = Invoke-GraphApi -Method Post -ApiPath oauth2PermissionGrants -Body (ConvertTo-Json ([pscustomobject]@{
            'odata.type' = 'Microsoft.DirectoryServices.OAuth2PermissionGrant'
            clientId     = $clientApplicationServicePrincipal.objectId
            resourceId   = $resourceApplicationServicePrincipal.objectId
            consentType  = $ConsentType
            scope        = $Scope
            startTime    = [DateTime]::UtcNow.ToString('o')
            expiryTime   = [DateTime]::UtcNow.AddYears(1).ToString('o')
        }))

        Write-Verbose "Sleeping for 3 seconds to allow the permission grant to propagate..." -Verbose
        Start-Sleep -Seconds 3
    }
    else
    {
        Write-Verbose "Existing OAuth2PermissionGrant found: $(ConvertTo-Json $existingGrant)" -Verbose

        $existingScopes = $existingGrant.scope.Split(' ')
        $missingScopes  = $scopesToGrant | Where { $_ -inotin $existingScopes }

        if ($missingScopes.Count)
        {
            $fullScopes = $existingGrant.scope += (' ' + [string]::Join(' ', $missingScopes))
            Write-Verbose "Updating OAuth2PermissionGrant scopes to include missing scopes '$([string]::Join(' ', $missingScopes))' to client application service principal '$($clientApplicationServicePrincipal.appDisplayName)' on behalf of resource application '$($resourceApplicationServicePrincipal.appDisplayName)'. Full Scopes will include: '$fullScopes'" -Verbose
            $response = Invoke-GraphApi -Method Patch -ApiPath "oauth2PermissionGrants/$($existingGrant.objectId)" -Body (ConvertTo-Json ([pscustomobject]@{ scope = $fullScopes }))
        }
        else
        {
            Write-Verbose "OAuth2Permission '$Scope' already granted to client application service principal '$($clientApplicationServicePrincipal.appDisplayName)' on behalf of resource application '$($resourceApplicationServicePrincipal.appDisplayName)'. Full Scopes: '$($existingGrant.scope)'" -Verbose
        }

    }
}

<#
.Synopsis
   Idempotently creates an application role assignment for an application service principal against another application service principal in Graph with the specified properties (service principals are created if they do not exist).
#>
function Initialize-GraphAppRoleAssignment
{
    [CmdletBinding(DefaultParameterSetName='ClientAppId_ResourceAppId')]
    param
    (
        # The application identifier of the client application.
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_ResourceAppId')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_ResourceIdentifierUri')]
        [ValidateNotNullOrEmpty()]
        [string] $ClientApplicationId,

        # The application identifier URI of the client application.
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_ResourceAppId')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_ResourceIdentifierUri')]
        [ValidateNotNullOrEmpty()]
        [string] $ClientApplicationIdentifierUri,

        # The application identifier of the resource application.
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_ResourceAppId')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_ResourceAppId')]
        [ValidateNotNullOrEmpty()]
        [string] $ResourceApplicationId,

        # The application identifier URI of the resource application.
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_ResourceIdentifierUri')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_ResourceIdentifierUri')]
        [ValidateNotNullOrEmpty()]
        [string] $ResourceApplicationIdentifierUri,

        # The identifier of the app role permission to grant.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $RoleId,

        # The type of the service principal to with the permission will be granted (e.g. 'ServicePrincipal' [default value]).
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [ValidateSet('ServicePrincipal', 'User', 'Group')]
        [string] $PrincipalType = 'ServicePrincipal'
    )

    # https://msdn.microsoft.com/en-us/library/azure/ad/graph/api/entity-and-complex-type-reference#approleassignment-entity

    # Ensure the application service principals exist in the directory tenant

    $clientApplicationServicePrincipal = if ($ClientApplicationId)
    {
        Initialize-GraphApplicationServicePrincipal -ApplicationId $ClientApplicationId -ErrorAction Stop
    }
    else
    {
        Get-GraphApplicationServicePrincipal -ApplicationIdentifierUri $ClientApplicationIdentifierUri -ErrorAction Stop
    }

    $resourceApplicationServicePrincipal = if ($ResourceApplicationId)
    {
        Initialize-GraphApplicationServicePrincipal -ApplicationId $ResourceApplicationId -ErrorAction Stop
    }
    else
    {
        Get-GraphApplicationServicePrincipal -ApplicationIdentifierUri $ResourceApplicationIdentifierUri -ErrorAction Stop
    }

    $existingAssignments = (Invoke-GraphApi -ApiPath "servicePrincipals/$($clientApplicationServicePrincipal.objectId)/appRoleAssignedTo").value
    $existingAssignment  = $existingAssignments |
        Where id -EQ $RoleId |
        Where resourceId -EQ $resourceApplicationServicePrincipal.objectId

    if (-not $existingAssignment)
    {
        Write-Verbose "Granting AppRoleAssignment '$RoleId' to application service principal '$($clientApplicationServicePrincipal.appDisplayName)' on behalf of application '$($resourceApplicationServicePrincipal.appDisplayName)'..." -Verbose
        $response = Invoke-GraphApi -Method Post -ApiPath "servicePrincipals/$($clientApplicationServicePrincipal.objectId)/appRoleAssignments" -Body (ConvertTo-Json ([pscustomobject]@{
            principalId   = $clientApplicationServicePrincipal.objectId
            principalType = $PrincipalType
            resourceId    = $resourceApplicationServicePrincipal.objectId
            id            = $RoleId
        }))
    }
    else
    {
        Write-Verbose "AppRoleAssignment '$RoleId' already granted to client application service principal '$($clientApplicationServicePrincipal.appDisplayName)' on behalf of resource application '$($resourceApplicationServicePrincipal.appDisplayName)'." -Verbose
    }
}

<#
.Synopsis
   Idempotently grants an application service principal membership to a directory role to (service principal is created if it do not exist).
#>
function Initialize-GraphDirectoryRoleMembership
{
    [CmdletBinding()]
    param
    (
        # The application identifier to which role membership should be granted.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $ApplicationId,

        # The display name of the role to which the application should be granted membership.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [ValidateSet('Directory Readers')]
        [string] $RoleDisplayName
    )

    # Ensure the application service principal exists in the directory tenant
    $applicationServicePrincipal = Initialize-GraphApplicationServicePrincipal -ApplicationId $ApplicationId

    # https://msdn.microsoft.com/en-us/Library/Azure/Ad/Graph/api/directoryroles-operations#AddDirectoryRoleMembers

    # Lookup the object id of the directory role in the directory tenant (note - these reference role templates)
    $roles        = Invoke-GraphApi -ApiPath directoryRoles
    $roleObjectId = $roles.value | Where displayName -EQ $RoleDisplayName | Select -First 1 -ExpandProperty objectId
    Write-Verbose "Existing Directory Roles: $(ConvertTo-Json $roles.value)" -Verbose

    # If the directory readers role does not exist, we need to "activate it"
    if (-not $roleObjectId)
    {
        $roleTemplates  = Invoke-GraphApi -ApiPath directoryRoleTemplates
        $roleTemplateId = $roleTemplates.value | Where displayName -EQ $RoleDisplayName | Select -First 1 -ExpandProperty objectId
        Write-Verbose "Existing Directory Role Templates: $(ConvertTo-Json $roleTemplates.value)" -Verbose

        Write-Verbose "Creating directory role '$RoleDisplayName' ($($roleTemplateId))..." -Verbose
        $response = Invoke-GraphApi -Method Post -ApiPath directoryRoles -Body (ConvertTo-Json ([pscustomobject]@{
            roleTemplateId = $roleTemplateId
        }))

        $roleObjectId = $response.objectId
    }

    # Lookup the existing memberships of the service principal; if the application service principal is not already a member of the directory role, grant it role membership
    $apiPath  = "servicePrincipals/$($applicationServicePrincipal.objectId)/getMemberObjects"
    $response = Invoke-GraphApi -Method Post -ApiPath $apiPath -Body (ConvertTo-Json ([pscustomobject]@{
        securityEnabledOnly = $false
    }))

    if ($response.value -icontains $roleObjectId)
    {
        Write-Verbose "Membership already granted to directory role '$RoleDisplayName' ($($roleObjectId)) for application service principal '$($applicationServicePrincipal.appDisplayName)'." -Verbose
    }
    else
    {
        Write-Verbose "Granting membership to directory role '$RoleDisplayName' ($($roleObjectId)) for application service principal '$($applicationServicePrincipal.appDisplayName)'..." -Verbose
        $apiPath  = "directoryRoles/$roleObjectId/`$links/members"
        $response = Invoke-GraphApi -Method Post -ApiPath $apiPath -Body (ConvertTo-Json ([pscustomobject]@{
            url = '{0}/directoryObjects/{1}' -f $Script:GraphEnvironment.GraphEndpoint.AbsoluteUri.TrimEnd('/'), $applicationServicePrincipal.objectId
        }))
    }
}

<#
.Synopsis
   Creates a new representation of a permission exposed by a resource application and grantable to a client application.
#>
function New-GraphPermissionDescription
{
    [CmdletBinding(DefaultParameterSetName='ClientAppId_ResourceAppId')]
    param
    (
        # The application identifier of the client application.
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_ResourceAppId')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_ResourceIdentifierUri')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_Resource')]
        [ValidateNotNullOrEmpty()]
        [string] $ClientApplicationId,

        # The application identifier URI of the client application.
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_ResourceAppId')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_ResourceIdentifierUri')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_Resource')]
        [ValidateNotNullOrEmpty()]
        [string] $ClientApplicationIdentifierUri,

        # The object reprsentation client application service principal.
        [Parameter(Mandatory=$true, ParameterSetName='Client_ResourceAppId')]
        [Parameter(Mandatory=$true, ParameterSetName='Client_ResourceIdentifierUri')]
        [Parameter(Mandatory=$true, ParameterSetName='Client_Resource')]
        [pscustomobject] $ClientApplicationServicePrincipal,

        # The application identifier of the resource application.
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_ResourceAppId')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_ResourceAppId')]
        [Parameter(Mandatory=$true, ParameterSetName='Client_ResourceAppId')]
        [ValidateNotNullOrEmpty()]
        [string] $ResourceApplicationId,

        # The application identifier URI of the resource application.
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_ResourceIdentifierUri')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_ResourceIdentifierUri')]
        [Parameter(Mandatory=$true, ParameterSetName='Client_ResourceIdentifierUri')]
        [ValidateNotNullOrEmpty()]
        [string] $ResourceApplicationIdentifierUri,

        # The object reprsentation resource application service principal.
        [Parameter(Mandatory=$true, ParameterSetName='ClientAppId_Resource')]
        [Parameter(Mandatory=$true, ParameterSetName='ClientIdentifierUri_Resource')]
        [Parameter(Mandatory=$true, ParameterSetName='Client_Resource')]
        [pscustomobject] $ResourceApplicationServicePrincipal,

        # The type of the permission.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [ValidateSet('Application', 'Delegated')]
        [string] $PermissionType,

        # The identifier of the permission.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $PermissionId,

        # Indicates whether the permission has been granted (consented).
        [Parameter()]
        [ValidateNotNull()]
        [switch] $IsConsented,

        # Indicates whether the current permission consent status should be queried.
        [Parameter()]
        [ValidateNotNull()]
        [switch] $LookupConsentStatus
    )

    # Lookup / initialize client service principal
    $ClientApplicationServicePrincipal = if ($ClientApplicationServicePrincipal)
    {
        $ClientApplicationServicePrincipal
    }
    elseif ($ClientApplicationId)
    {
        Initialize-GraphApplicationServicePrincipal -ApplicationId $ClientApplicationId -ErrorAction Stop
    }
    else
    {
        Get-GraphApplicationServicePrincipal -ApplicationIdentifierUri $ClientApplicationIdentifierUri -ErrorAction Stop
    }

    # Lookup / initialize resource service principal
    $ResourceApplicationServicePrincipal = if ($ResourceApplicationServicePrincipal)
    {
        $ResourceApplicationServicePrincipal
    }
    elseif ($ResourceApplicationId)
    {
        Initialize-GraphApplicationServicePrincipal -ApplicationId $ResourceApplicationId -ErrorAction Stop
    }
    else
    {
        Get-GraphApplicationServicePrincipal -ApplicationIdentifierUri $ResourceApplicationIdentifierUri -ErrorAction Stop
    }

    $permissionProperties = [ordered]@{
        clientApplicationId            = $ClientApplicationServicePrincipal.appId
        clientApplicationDisplayName   = $ClientApplicationServicePrincipal.appDisplayName
        resourceApplicationId          = $ResourceApplicationServicePrincipal.appId
        resourceApplicationDisplayName = $ResourceApplicationServicePrincipal.appDisplayName
    }

    $permissionProperties += [ordered]@{
        isConsented    = $IsConsented
        permissionType = $PermissionType
        permissionId   = $PermissionId
    }

    switch ($PermissionType)
    {
        'Application'
        {
            $appRole = $ResourceApplicationServicePrincipal.appRoles | Where id -EQ $PermissionId
            $permissionProperties += [ordered]@{
                permissionName        = $appRole.value
                permissionDisplayName = $appRole.displayName
                permissionDescription = $appRole.description
            }

            if ($LookupConsentStatus)
            {
                $existingAppRoleAssignments = (Invoke-GraphApi -ApiPath "servicePrincipals/$($ClientApplicationServicePrincipal.objectId)/appRoleAssignedTo").value
                $permissionProperties.isConsented = if ($existingAppRoleAssignments | Where id -EQ $PermissionId) {$true} else {$false}
            }
        }

        'Delegated'
        {
            $oAuth2Permission = $ResourceApplicationServicePrincipal.oauth2Permissions | Where id -EQ $PermissionId
            $permissionProperties += [ordered]@{
                permissionName        = $oAuth2Permission.value
                permissionDisplayName = $oAuth2Permission.adminConsentDisplayName
                permissionDescription = $oAuth2Permission.adminConsentDescription
            }

            if ($LookupConsentStatus)
            {
                $queryParameters = @{
                    '$filter' = "resourceId eq '$($ResourceApplicationServicePrincipal.objectId)' and clientId eq '$($ClientApplicationServicePrincipal.objectId)'"
                    '$top'    = '500' # Note - there is an issue with this API if you use a large page size and the result set is large as well
                }
                $existingOAuth2PermissionGrants = (Invoke-GraphApi -ApiPath oauth2PermissionGrants -QueryParameters $queryParameters).Value
                $permissionProperties.isConsented = if ($existingOAuth2PermissionGrants) {$true} else {$false}
            }
        }
    }

    Write-Output ([pscustomobject]$permissionProperties)
}

<#
.Synopsis
   Gets all permissions which have been granted to the specified application. If the application was created in the current directory tenant, also returns permissions which have not been consented but which are advertised as "required" in the application's manifest.
#>
function Get-GraphApplicationPermissions
{
    [CmdletBinding()]
    [OutputType([pscustomobject])]
    param
    (
        # The application identifier for which all consented permissions should be retrieved.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [Alias('appId')]
        [string] $ApplicationId
    )

    # Ensure the application service principal exists in the directory tenant
    $applicationServicePrincipal = Initialize-GraphApplicationServicePrincipal -ApplicationId $ApplicationId -ErrorAction Stop

    # Identify which permissions have already been granted
    $existingAppRoleAssignments = (Invoke-GraphApi -ApiPath "servicePrincipals/$($applicationServicePrincipal.objectId)/appRoleAssignedTo" -ErrorAction Stop).value
    
    # Note - there is an issue with the oauth2PermissionGrants API when using an OData filter querying for a resourceId; if the resulting collection is greater than 1000 members, the API call will return with status 200 but the response will include a malformed JSON odata error message
    # Note - there is an issue with the oauth2PermissionGrants API; any client which has more than 100 permissions will only have the first 100 permissions returned
    # Note - there is an issue with the oauth2PermissionGrants API if you use a large page size and the result set is large as well
    $existingClientOAuth2PermissionGrants = @()
    try
    {
        $existingClientOAuth2PermissionGrants = (Invoke-GraphApi -ApiPath oauth2PermissionGrants -QueryParameters @{ '$filter' = "clientId eq '$($applicationServicePrincipal.objectId)'"; '$top' = '500' } -ErrorAction Stop).Value
    }
    catch
    {
        Write-Warning "An issue occurred trying to lookup OAuth2PermissionGrants where application '$ApplicationId' is a client; Omitting this class of permissions from resulting data; Error: $_"
    }

    $existingResourceOAuth2PermissionGrants = @()
    try
    {
        $existingResourceOAuth2PermissionGrants = (Invoke-GraphApi -ApiPath oauth2PermissionGrants -QueryParameters @{ '$filter' = "resourceId eq '$($applicationServicePrincipal.objectId)'"; '$top' = '500' } -ErrorAction Stop).Value
    }
    catch
    {
        Write-Warning "An issue occurred trying to lookup OAuth2PermissionGrants where application '$ApplicationId' is a resource; Omitting this class of permissions from resulting data; Error: $_"
    }

    # Build a representation of each permission which has been granted
    $permissions = @()
    foreach ($existingAppRoleAssignment in $existingAppRoleAssignments)
    {
        $permissionParams = @{
            ClientApplicationServicePrincipal   = $applicationServicePrincipal
            ResourceApplicationServicePrincipal = Invoke-GraphApi -ApiPath "directoryObjects/$($existingAppRoleAssignment.resourceId)" -ErrorAction Stop
            PermissionType                      = 'Application'
            PermissionId                        = $existingAppRoleAssignment.id
            IsConsented                         = $true
        }
        $permissions += New-GraphPermissionDescription @permissionParams
    }
    foreach ($existingOAuth2PermissionGrant in $existingClientOAuth2PermissionGrants)
    {
        $permissionParams = @{
            ClientApplicationServicePrincipal   = $applicationServicePrincipal
            ResourceApplicationServicePrincipal = Invoke-GraphApi -ApiPath "directoryObjects/$($existingOAuth2PermissionGrant.resourceId)" -ErrorAction Stop
            PermissionType                      = 'Delegated'
            IsConsented                         = $true
        }
        foreach ($scope in $existingOAuth2PermissionGrant.scope.split(' '))
        {
            $oAuth2Permission = $permissionParams.ResourceApplicationServicePrincipal.oauth2Permissions | Where value -EQ $scope
            $permissions += New-GraphPermissionDescription @permissionParams -PermissionId $oAuth2Permission.id
        }
    }
    foreach ($existingOAuth2PermissionGrant in $existingResourceOAuth2PermissionGrants)
    {
        $permissionParams = @{
            ClientApplicationServicePrincipal   = Invoke-GraphApi -ApiPath "directoryObjects/$($existingOAuth2PermissionGrant.clientId)" -ErrorAction Stop
            ResourceApplicationServicePrincipal = $applicationServicePrincipal
            PermissionType                      = 'Delegated'
            IsConsented                         = $true
        }
        foreach ($scope in $existingOAuth2PermissionGrant.scope.split(' '))
        {
            $oAuth2Permission = $permissionParams.ResourceApplicationServicePrincipal.oauth2Permissions | Where value -EQ $scope
            $permissions += New-GraphPermissionDescription @permissionParams -PermissionId $oAuth2Permission.id
        }
    }

    # Attempt to get unconsented permissions if we can access the application object (e.g. if the application exists in the same directory in which we are currently authenticated)
    if (($application = Find-GraphApplication -AppId $ApplicationId))
    {
        foreach ($requiredResource in $application.requiredResourceAccess)
        {
            $permissionParams = @{
                ClientApplicationServicePrincipal   = $applicationServicePrincipal
                ResourceApplicationServicePrincipal = Initialize-GraphApplicationServicePrincipal -ApplicationId $requiredResource.resourceAppId
                IsConsented                         = $false
            }
            foreach ($resourceAccess in $requiredResource.resourceAccess)
            {
                $relatedConsentedPermissions = $permissions | Where resourceApplicationId -EQ $requiredResource.resourceAppId | Where permissionId -EQ $resourceAccess.id
                # $resourceAccess.type is one of: 'Role', 'Scope', 'Role,Scope', or 'Scope,Role'
                if ($resourceAccess.type -ilike '*Role*')
                {
                    if (-not ($relatedConsentedPermissions | Where permissionType -EQ 'Application'))
                    {
                        $permissions += New-GraphPermissionDescription @permissionParams -PermissionType Application -PermissionId $resourceAccess.id
                    }
                    else
                    {
                        Write-Verbose "Application permission '$($resourceAccess.id)' of type 'AppRoleAssignment' already consented for application '$($applicationServicePrincipal.appDisplayName)' ('$ApplicationId')."
                    }
                }
                if ($resourceAccess.type -ilike '*Scope*')
                {
                    if (-not ($relatedConsentedPermissions | Where permissionType -EQ 'Delegated'))
                    {
                        $permissions += New-GraphPermissionDescription @permissionParams -PermissionType Delegated -PermissionId $resourceAccess.id
                    }
                    else
                    {
                        Write-Verbose "Application permission '$($resourceAccess.id)' of type 'OAuth2PermissionGrant' already consented for application '$($applicationServicePrincipal.appDisplayName)' ('$ApplicationId')."
                    }
                }
            }
        }
    }
    else
    {
        Write-Verbose "Unable to retrieve application with appId '$ApplicationId' and will be unable to retrieve information on any additional required permissions which have not been consented." -Verbose
    }

    if (-not $permissions.Count)
    {
        if ($application)
        {
            Write-Verbose "Application '$($applicationServicePrincipal.appDisplayName)' ('$ApplicationId') does not have any consented permissions, nor does it advertise any additional required permissions in its application manifest." -Verbose
        }
        else
        {
            Write-Verbose "Application '$($applicationServicePrincipal.appDisplayName)' ('$ApplicationId') does not have any consented permissions in the current directory tenant." -Verbose
        }
    }

    Write-Output $permissions
}

<#
.Synopsis
   Grants a permission to a graph application. Use the 'New-GraphApplicationPermission' or 'Get-GraphApplicationPermissions' cmdlets to create an instance of the permission object or to see its structure.
#>
function Grant-GraphApplicationPermission
{
    [CmdletBinding()]
    param
    (
        # The graph permission description object.
        [Parameter(Mandatory=$true, ValueFromPipeline=$true)]
        [ValidateNotNullOrEmpty()]
        [pscustomobject] $PermissionDescription
    )
    process
    {
        Write-Verbose "Granting permission '$($PermissionDescription.permissionName)' ($($PermissionDescription.PermissionId)) exposed by application '$($PermissionDescription.resourceApplicationDisplayName)' ($($PermissionDescription.resourceApplicationId)) of type '$($PermissionDescription.PermissionType)' to application '$($PermissionDescription.clientApplicationDisplayName)' ($($PermissionDescription.clientApplicationId))" -Verbose
        $params = @{ ClientApplicationId = $PermissionDescription.clientApplicationId; ResourceApplicationId = $PermissionDescription.resourceApplicationId }
        switch ($PermissionDescription.permissionType)
        {
            'Application' { Initialize-GraphAppRoleAssignment     @params -RoleId $PermissionDescription.permissionId   -Verbose }
            'Delegated'   { Initialize-GraphOAuth2PermissionGrant @params -Scope  $PermissionDescription.permissionName -Verbose }
        }
    }
}

<#
.Synopsis
   Grants all permissions required by an application which are specified in the application manifest. Only applies to the home directory of the application.
#>
function Grant-GraphApplicationPermissions
{
    [CmdletBinding()]
    param
    (
        # The application identifier for which all required permissions should be granted.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $ApplicationId
    )
    # Ensure the application can be retrieved in the current directory tenant
    $application = Get-GraphApplication -AppId $ApplicationId -ErrorAction Stop
    $permissions = Get-GraphApplicationPermissions -ApplicationId $ApplicationId

    # Optimization / workaround for AzureChinaCloud where PATCH OAuth2PermissionGrant API has some issues
    # To mitigate we group into one permission grant with all scopes in a single POST call
    if ((Get-GraphEnvironmentInfo).Environment -eq 'AzureChinaCloud')
    {
        Write-Verbose "Grouping OAuth2PermissionGrants to avoid PATCH call..." -Verbose
        $groupedPermissions  = @($permissions | Where permissionType -EQ Application)
        $groupedPermissions += @($permissions | Where permissionType -EQ Delegated |
            Group clientApplicationId | ForEach { $_.Group | Group resourceApplicationId | ForEach {
                $params = @{
                    ClientApplicationId   = $_.Group[0].ClientApplicationId
                    ResourceApplicationId = $_.Group[0].ResourceApplicationId
                    PermissionType        = $_.Group[0].PermissionType
                    PermissionId          = $_.Group.PermissionName -join ' '
                    IsConsented           = $false
                    LookupConsentStatus   = $false
                }
                $permission = New-GraphPermissionDescription @params
                $permission.permissionName        = $permission.PermissionId
                $permission.permissionDisplayName = 'Aggregate Permission'
                $permission.permissionDescription = 'Aggregate Permission'
                Write-Output $permission
            }
        })
        $permissions = $groupedPermissions | Where {$_}
    }

    foreach ($permission in $permissions)
    {
        if ($permission.isConsented)
        {
            Write-Verbose "Permission '$($permission.permissionName)' ($($permission.PermissionId)) exposed by application '$($permission.resourceApplicationDisplayName)' ($($permission.resourceApplicationId)) of type '$($permission.PermissionType)' has already been granted to application '$($permission.clientApplicationDisplayName)' ($($permission.clientApplicationId))" -Verbose
        }
        else
        {
            Grant-GraphApplicationPermission -PermissionDescription $permission
        }
    }
}

<#
.Synopsis
   Writes a representation of the specified Graph permission descriptions to the current PowerShell host console window.
#>
function Show-GraphApplicationPermissionDescriptions
{
    [CmdletBinding()]
    param
    (
        # The graph permission description object.
        [Parameter(Mandatory=$true, ValueFromPipeline=$true)]
        [ValidateNotNullOrEmpty()]
        [pscustomobject[]] $PermissionDescription,

        # The text display to use above the permission descriptions.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [string] $DisplayHeader = 'Microsoft Azure Stack - Required Directory Permissions',

        # The text display to use below the permission descriptions.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [string] $DisplayFooter = '(X) = Permission granted, ( ) = Permission not granted',

        # Indicates that any duplicate permissions should be filtered-out from the display.
        [Parameter()]
        [Switch] $FilterDuplicates = $true
    )
    begin
    {
        $permissions = @()
    }
    process
    {
        foreach ($permission in $PermissionDescription)
        {
            if ($FilterDuplicates -and ($permissions |
                    Where clientApplicationId   -EQ $permission.clientApplicationId |
                    Where resourceApplicationId -EQ $permission.resourceApplicationId |
                    Where permissionId          -EQ $permission.permissionId |
                    Where permissionType        -EQ $permission.permissionType))
            {
                continue
            }

            $permissions += $permission
        }
    }
    end
    {
        <# Writes a textual consent display to the console similar to this:
        +--------------------------------------------------------+
        | Microsoft Azure Stack - Required Directory Permissions |
        +--------------------------------------------------------+
        | Access Azure Stack                                     |
        |   (X) Delegated to: Microsoft Azure Stack              |
        |                                                        |
        | Access the directory as the signed-in user             |
        |   (X) Delegated to: Azure Stack                        |
        |                                                        |
        | Read all users' basic profiles                         |
        |   (X) Delegated to: Azure Stack                        |
        |                                                        |
        | Read all users' full profiles                          |
        |   (X) Delegated to: Azure Stack                        |
        |                                                        |
        | Read directory data                                    |
        |   (X) Granted to:   Azure Stack                        |
        |   (X) Granted to:   Azure Stack - Policy               |
        |   (X) Delegated to: Azure Stack                        |
        |   (X) Delegated to: Microsoft Azure Stack              |
        |                                                        |
        | Sign in and read user profile                          |
        |   (X) Delegated to: Azure Stack                        |
        |   (X) Delegated to: Microsoft Azure Stack              |
        |                                                        |
        +--------------------------------------------------------+
        | (X) = Permission granted, ( ) = Permission not granted |
        +--------------------------------------------------------+
        #>

        $header = $DisplayHeader
        $footer = $DisplayFooter

        $lines = @()
        foreach ($permissionGroup in @($permissions | Sort resourceApplicationDisplayName, permissionDisplayName | Group permissionId))
        {
            $lines += "{0}" -f $permissionGroup.Group[0].permissionDisplayName
            foreach ($permission in @($permissionGroup.Group | Sort permissionType, clientApplicationDisplayName))
            {
                $lines += "  {0} {1} {2}" -f @(
                    ($consentDisplay = if ($permission.isConsented) {'(X)'} else {'( )'})
                    ($typeDisplay = switch ($permission.permissionType) { 'Application' { 'Granted to:  ' }; 'Delegated' { 'Delegated to:' } })
                    $permission.clientApplicationDisplayName
                )
            }
            $lines += ''
        }

        $max = (($lines + @($header, $footer)) | Measure Length -Maximum).Maximum
        $div = '+-{0}-+' -f (New-Object string('-', $max))

        $lines = @(
            $div
            '| {0} |' -f "$header".PadRight($max)
            $div
            $lines | ForEach { '| {0} |' -f "$_".PadRight($max) }
            $div
            '| {0} |' -f "$footer".PadRight($max)
            $div
        )

        foreach ($line in $lines)
        {
            Write-Host $line
        }
    }
}

<#
.Synopsis
   Creates or updates an application in Graph with an implicit service principal and the specified properties.
#>
function Initialize-GraphApplication
{
    [CmdletBinding(DefaultParameterSetName='Cert')]
    [OutputType([pscustomobject])]
    param
    (
        # The display name of the application.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $DisplayName,

        # The homepage address of the application.
        [Parameter(Mandatory=$false)]
        [ValidateNotNullOrEmpty()]
        [string] $Homepage,

        # The reply address(es) of the application.
        [Parameter(Mandatory=$false)]
        [ValidateNotNullOrEmpty()]
        [string[]] $ReplyAddress,

        # The application identifier URI.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $IdentifierUri,

        # The client certificates used to authenticate with graph as the application / service principal.
        [Parameter(ParameterSetName='Cert')]
        [ValidateNotNull()]
        [System.Security.Cryptography.X509Certificates.X509Certificate2[]] $ClientCertificate = @(),

        # The thumbprint of the client certificate used to authenticate with graph as the application / service principal.
        [Parameter(ParameterSetName='Thumbprint')]
        [ValidateNotNull()]
        [ValidatePattern('^([0-9A-Fa-f]{2})*$')]
        [string] $ClientCertificateThumbprint = $null,

        # The set of AAD permissions required directly by the application in the context of its service principal.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [ValidateSet(
            'ReadDirectoryData',
            'ManageAppsThatThisAppCreatesOrOwns'
        )]
        [String[]] $ApplicationAadPermissions = @(),

        # The set of delegated AAD permissions required by the application in the context of the signed-in user.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [ValidateSet(
            'AccessDirectoryAsSignedInUser',
            'EnableSignOnAndReadUserProfiles',
            'ReadAllGroups',
            'ReadAllUsersBasicProfile',
            'ReadAllUsersFullProfile',
            'ReadDirectoryData'
        )]
        [String[]] $DelegatedAadPermissions = @(),

        # A collection of Application Identifier URIs to which delegated resource access (on behalf of the signed-in user - "user_impersonation") is required by / should be granted to the initialized application.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [String[]] $ResourceAccessByAppUris = @(),

        # The first-party applications to which the "user_impersonation" OAuth2 permission should be granted.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [ValidateSet(
            'LegacyPowerShell',
            'PowerShell',
            'VisualStudio',
            'AzureCLI'
        )]
        [string[]] $OAuth2PermissionGrants = @(),

        # A collection of Application Identifier URIs for which a service principal should be initialized and to which the "user_impersonation" OAuth2 permission should be granted.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [string[]] $OAuth2PermissionGrantsByAppUris = @(),

        # A collection of Application Identifier URIs for known client applications which should be associated to this application.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [string[]] $KnownClientApplicationsByAppUris = @(),

        # Tags to include in the application service principal.
        [Parameter()]
        [ValidateNotNullOrEmpty()]
        [string[]] $Tags = @(),

        # Indicates that the application should be deleted and re-created (if it already exists).
        [Parameter()]
        [Switch] $DeleteAndCreateNew,

        # Indicates that the application should be available to other tenants (multi-tenanted). True by default.
        [Parameter()]
        [Switch] $AvailableToOtherTenants = $true,

        # Indicates that the application service principal should have all declared application permissions consented-to. True by default.
        [Parameter()]
        [Switch] $ConsentToAppPermissions = $true,

        # Indicates that any existing client certificates associated to this application should be removed. False by default.
        [Parameter()]
        [Switch] $RemoveExistingClientCertificates
    )

    if ($ClientCertificateThumbprint)
    {
        $ClientCertificate += Get-Item "Cert:\LocalMachine\My\$ClientCertificateThumbprint" -ErrorAction Stop
    }

    if (($existingApplication = Find-GraphApplication -AppUri $IdentifierUri -ErrorAction Stop))
    {
        if ($DeleteAndCreateNew)
        {
            # Very special case of updating multi-tenanted application before removing them
            if ($existingApplication.availableToOtherTenants)
            {
                Write-Verbose "Disable multi-tenancy before removing the application..." -Verbose
                $existingApplication.availableToOtherTenants = $false
                $apiPath = "directoryObjects/$($existingApplication.objectId)/Microsoft.DirectoryServices.Application"
                $requestBodyAsJson =  @{ availableToOtherTenants = $false } | ConvertTo-Json -Depth 10
                $noContentResponse = Invoke-GraphApi -Method Patch -ApiPath $apiPath -Body $requestBodyAsJson -ErrorAction Stop
            }

            Write-Warning "Existing application identified by '$IdentifierUri' with id '$($existingApplication.ObjectId)' found. Deleting application..."
            $existingApplication | Remove-GraphObject -ErrorAction Stop
            $existingApplication = $null

            while (Find-GraphApplication -AppUri $IdentifierUri -ErrorAction Stop -Verbose)
            {
                Write-Verbose "Waiting for graph application to be deleted..." -Verbose
                Start-Sleep -Seconds 1
            }
        }
        else
        {
            Write-Verbose "An existing application with identifier '$IdentifierUri' was found. This application will be updated accordingly." -Verbose
        }
    }
    else
    {
        Write-Verbose "Existing application with identifier '$IdentifierUri' not found. A new one will be created." -Verbose
    }

    ##
    ##

    # Initialize the request body
    $requestBody = @{
        "odata.type"          = "Microsoft.DirectoryServices.Application"
        displayName           = $DisplayName
        groupMembershipClaims = "SecurityGroup" # Note: Possible values are "null" => means No Claims, "SecurityGroup" => means 'SG and Azure AD roles' and "All" => means "SG + DL + Azure AD roles"

        # Initialize the application identifiers, preserving any that already exist, and idempotently adding the specified URI into the collection
        "identifierUris@odata.type" = "Collection(Edm.String)"
        identifierUris              = @(@($existingApplication.identifierUris) + @($IdentifierUri) | Select -Unique | Where { $_ -ne $null })
    }

    # Enable multi-tenancy if applicable
    if ($AvailableToOtherTenants)
    {
        if ("$($existingApplication.availableToOtherTenants)" -ieq 'false')
        {
            Write-Warning "Existing application with identifier '$IdentifierUri' was previously created with configuration 'availableToOtherTenants = false'. Updating configuration to 'availableToOtherTenants = true'." -ErrorAction Stop
        }

        $requestBody += @{
            availableToOtherTenants = $true
        }
    }

    # Initialize the application reply URLs, preserving any that already exist, and idempotently adding the specified URI into the collection
    if ($ReplyAddress.Count -gt 0)
    {
        $requestBody += @{
            "replyUrls@odata.type" = "Collection(Edm.String)"
            replyUrls              = @(@($existingApplication.replyUrls) + $ReplyAddress | Select -Unique | Where { $_ -ne $null })
        }
    }

    if ($Homepage)
    {
        $requestBody['homepage'] = $Homepage
    }

    # Initialize the application key credentials with which it can authenticate
    $requestBody['keyCredentials@odata.type'] = "Collection(Microsoft.DirectoryServices.KeyCredential)"
    $requestBody['keyCredentials'] = @(@($existingApplication.keyCredentials) | Where { $_ -ne $null })
    if ($RemoveExistingClientCertificates)
    {
        $requestBody['keyCredentials'] = @()
    }
    foreach ($cert in $ClientCertificate)
    {
        $customKeyIdentifier = [Convert]::ToBase64String($cert.GetCertHash())
        if (-not (@($requestBody['keyCredentials']) | Where customKeyIdentifier -EQ $customKeyIdentifier))
        {
            Write-Verbose "Adding new key credentials to application using client certificate '$($cert.Subject)' ($($cert.Thumbprint))" -Verbose

            $requestBody['keyCredentials'] += @(,([pscustomobject]@{
                keyId               = [Guid]::NewGuid()
                type                = "AsymmetricX509Cert"
                usage               = "Verify"
                customKeyIdentifier = $customKeyIdentifier
                value               = [Convert]::ToBase64String($cert.GetRawCertData())
                startDate           = $cert.NotBefore.ToUniversalTime().ToString('o')
                endDate             = $cert.NotAfter.ToUniversalTime().ToString('o')
            }))
        }
        else
        {
            Write-Verbose "Key credentials already exist on application for client certificate '$($cert.Subject)' ($($cert.Thumbprint))" -Verbose
        }
    }

    # Initialize required AAD permissions
    $aadPermissions   = @()
    $rolePermissions  = New-Object System.Collections.Generic.HashSet[string](,[string[]]$ApplicationAadPermissions)
    $scopePermissions = New-Object System.Collections.Generic.HashSet[string](,[string[]]$DelegatedAadPermissions)
    $allPermissions   = New-Object System.Collections.Generic.HashSet[string](,[string[]]($rolePermissions + $scopePermissions))
    foreach ($permissionName in $allPermissions)
    {
        $permissionType = ''
        if ($rolePermissions.Contains($permissionName))  { $permissionType += 'Role,' }
        if ($scopePermissions.Contains($permissionName)) { $permissionType += 'Scope' }

        $aadPermissions += [pscustomobject]@{
            id   = $Script:GraphEnvironment.AadPermissions[$permissionName]
            type = $permissionType.Trim(',')
        }
    }

    if ($aadPermissions.Count -gt 0)
    {
        if (-not ($existingRequiredResourceAccess = @($existingApplication.requiredResourceAccess) | Where resourceAppId -EQ $Script:GraphEnvironment.Applications.WindowsAzureActiveDirectory.Id))
        {
            $existingRequiredResourceAccess = @{
                "resourceAccess@odata.type" = "Collection(Microsoft.DirectoryServices.ResourceAccess)"
                resourceAppId  = $Script:GraphEnvironment.Applications.WindowsAzureActiveDirectory.Id
                resourceAccess = @()
            }

            if (-not $requestBody['requiredResourceAccess'])
            {
                $requestBody['requiredResourceAccess@odata.type'] = "Collection(Microsoft.DirectoryServices.RequiredResourceAccess)"
                $requestBody['requiredResourceAccess'] = @(@($existingApplication.requiredResourceAccess) | Where { $_ -ne $null })
            }

            $requestBody['requiredResourceAccess'] += ,$existingRequiredResourceAccess
        }

        foreach ($aadPermission in $aadPermissions)
        {
            if (-not ($existingRequiredResourceAccess.resourceAccess | Where id -EQ $aadPermission.id))
            {
                Write-Verbose "Adding permission ($($aadPermission.id)) on AAD application ($($existingRequiredResourceAccess.resourceAppId))" -Verbose
                
                if (-not $requestBody['requiredResourceAccess'])
                {
                    $requestBody['requiredResourceAccess@odata.type'] = "Collection(Microsoft.DirectoryServices.RequiredResourceAccess)"
                    $requestBody['requiredResourceAccess'] = @(@($existingApplication.requiredResourceAccess) | Where { $_ -ne $null })
                }
                
                $existingRequiredResourceAccess.resourceAccess += ,$aadPermission
            }
            else
            {
                Write-Verbose "Permission ($($aadPermission.id)) already granted on AAD application ($($existingRequiredResourceAccess.resourceAppId))" -Verbose
            }
        }
    }

    # Initialize required permissions for other applications
    $permissionValue = 'user_impersonation'
    foreach ($appUri in $ResourceAccessByAppUris)
    {
        if (-not ($existingResourceApplication = Find-GraphApplication -AppUri $appUri))
        {
            Write-Error "Application '$appUri' does not exist. Unable to grant resource access for permission '$permissionValue' for this application to the target application."
            continue
        }

        if (-not ($existingRequiredResourceAccess = @($existingApplication.requiredResourceAccess) | Where resourceAppId -EQ $existingResourceApplication.appId))
        {
            $existingRequiredResourceAccess = @{
                "resourceAccess@odata.type" = "Collection(Microsoft.DirectoryServices.ResourceAccess)"
                resourceAppId  = $existingResourceApplication.appId
                resourceAccess = @()
            }

            if (-not $requestBody['requiredResourceAccess'])
            {
                $requestBody['requiredResourceAccess@odata.type'] = "Collection(Microsoft.DirectoryServices.RequiredResourceAccess)"
                $requestBody['requiredResourceAccess'] = @(@($existingApplication.requiredResourceAccess) | Where { $_ -ne $null })
            }

            $requestBody['requiredResourceAccess'] += ,$existingRequiredResourceAccess
        }

        if (-not ($permissionId = $existingResourceApplication.oauth2Permissions | Where Value -EQ $permissionValue | Select -First 1 -ExpandProperty Id))
        {
            Write-Error "OAuth2Permission for '$permissionValue' does not exist on application '$appUri' ($($existingResourceApplication.appId)) and cannot be granted to this application ($IdentifierUri)'."
            continue
        }

        if (-not ($existingRequiredResourceAccess.resourceAccess | Where id -EQ $permissionId))
        {
            Write-Verbose "Adding OAuth2 Permission for this application ('$($IdentifierUri)') to application '$appUri' ($($existingResourceApplication.appId))." -Verbose
            $existingRequiredResourceAccess.resourceAccess += ,@{
                id   = $permissionId
                type = "Scope"
            }
        }
        else
        {
            Write-Verbose "OAuth2 Permission for this application ('$($IdentifierUri)') already granted to application '$appUri' ($($existingResourceApplication.appId))." -Verbose
        }
    }

    # Initialize KnownClientApplications
    foreach ($appUri in $KnownClientApplicationsByAppUris)
    {
        if (-not ($clientApplication = Find-GraphApplication -AppUri $appUri))
        {
            Write-Error "Application '$appUri' does not exist. Unable to reference known client application relationship for this application to the target application."
            continue
        }

        if (-not $requestBody['knownClientApplications'])
        {
            $requestBody['knownClientApplications'] = @(@($existingApplication.knownClientApplications) | Where { $_ -ne $null })
        }

        if ($requestBody['knownClientApplications'] -inotcontains $clientApplication.appId)
        {
            Write-Verbose "Known client application '$appUri' ($($clientApplication.appId)) added to this application ('$($IdentifierUri)')" -Verbose
            $requestBody['knownClientApplications'] += $clientApplication.appId
        }
        else
        {
            Write-Verbose "Known client application '$appUri' ($($clientApplication.appId)) already added to this application ('$($IdentifierUri)')" -Verbose
        }
    }

    # Create or update the application
    $requestBodyAsJson = $requestBody | ConvertTo-Json -Depth 10

    if ($existingApplication)
    {
        Write-Verbose "Updating application in AAD..." -Verbose
        $apiPath = "directoryObjects/$($existingApplication.objectId)/Microsoft.DirectoryServices.Application"
        $noContentResponse = Invoke-GraphApi -Method Patch -ApiPath $apiPath -Body $requestBodyAsJson -ErrorAction Stop
        Start-Sleep -Seconds 5 # Delay between PATCH and GET to mitigate replication delay issues
        $application = Get-GraphApplication -AppUri $IdentifierUri
    }
    else
    {
        # Note: the post response does not always contain the accurate application state, so make a GET call to ensure it is accurate
        Write-Verbose "Creating application in AAD..." -Verbose
        $inaccurateResponse = Invoke-GraphApi -Method Post -ApiPath 'applications' -Body $requestBodyAsJson -ErrorAction Stop
        $application = Get-GraphObjectWithRetry -GetScript {Find-GraphApplication -AppUri $IdentifierUri} -MaxAttempts 10 -DelayInSecondsBetweenAttempts 5 -MinimumDelayInSeconds 5
    }

    # If the application does not have the user_impersonation permission, update it to include this permission
    # Note: this is a workaround to address the behavior in AzureChinaCloud which does not automatically include this permission
    if (-not $application.oauth2Permissions.value -icontains 'user_impersonation')
    {
        $requestBody['oauth2Permissions'] += @([pscustomobject]@{
            adminConsentDescription = "Allow the application to access $($application.DisplayName) on behalf of the signed-in user."
            adminConsentDisplayName = "Access $($application.DisplayName)"
            id                      = [guid]::NewGuid().ToString()
            isEnabled               = $true
            type                    = 'User'
            userConsentDescription  = "Allow the application to access $($application.DisplayName) on your behalf."
            userConsentDisplayName  = "Access $($application.DisplayName)"
            value                   = 'user_impersonation'
        })

        # Note: we must exclude the key credentials from the patch request, or null-out the key values. I haven't tried all possible combinations of which properties must be included wich cant be omitted, but I cannot just send the patch request with the update OAuth2Permissions.
        # {"odata.error":{"code":"Request_BadRequest","message":{"lang":"en","value":"Existing credential with KeyId '0d330a36-d042-41d9-b4bb-cdbb26be0595' must be sent back with null value."},"values":[{"item":"PropertyName","value":"keyCredentials"},{"item":"PropertyErrorCode","value":"KeyValueMustBeNull"}]}}
        $requestBody.Remove('keyCredentials')
        $requestBody.Remove('keyCredentials@odata.type')
        $patchRequestBodyAsJson = $requestBody | ConvertTo-Json -Depth 10

        Write-Warning "Application does not include the oauth2permission 'user_impersonation'! Updating application to include this permission..."
        $apiPath = "directoryObjects/$($application.objectId)/Microsoft.DirectoryServices.Application"
        $noContentResponse = Invoke-GraphApi -Method Patch -ApiPath $apiPath -Body $patchRequestBodyAsJson -ErrorAction Stop
        
		Start-Sleep -Seconds 5 # Delay between PATCH and GET to mitigate replication delay issues

        $application = Get-GraphApplication -AppUri $IdentifierUri
    }

    # Create a service principal for the application (if one doesn't already exist)
    $primaryServicePrincipal = Initialize-GraphApplicationServicePrincipal -ApplicationId $application.appId -Tags $Tags

    # Initialize OAuth2Permission grants to other (first-party) applications
    foreach ($applicationName in $OAuth2PermissionGrants)
    {
        $params = @{
            ClientApplicationId   = $Script:GraphEnvironment.Applications."$applicationName".Id
            ResourceApplicationId = $application.appId
        }

        Initialize-GraphOAuth2PermissionGrant @params
    }

    # Initialize OAuth2Permission grants to other (non-first-party) applications
    foreach ($applicationUri in $OAuth2PermissionGrantsByAppUris)
    {
        if (-not ($targetApplication = Find-GraphApplication -AppUri $applicationUri))
        {
            Write-Error "Application '$applicationUri' does not exist. Unable to grant OAuth2Permissions for this application to the target application."
            continue
        }

        $params = @{
            ClientApplicationId   = $targetApplication.appId
            ResourceApplicationId = $application.appId
        }

        Initialize-GraphOAuth2PermissionGrant @params
    }

    # "Consent" to application permissions
    if ($ConsentToAppPermissions)
    {
        Grant-GraphApplicationPermissions -ApplicationId $application.appId
    }

    # Return the application in its final (current) state
    Get-GraphApplication -AppUri $IdentifierUri | Write-Output
}

<#
.Synopsis
   Creates or updates an application in Graph with an implicit service principal and the specified properties.
#>
function Initialize-GraphApplicationOwner
{
    [CmdletBinding(DefaultParameterSetName='ById')]
    param
    (
        # The application identifier.
        [Parameter(Mandatory=$true, ParameterSetName='ById')]
        [ValidateNotNullOrEmpty()]
        [string] $ApplicationId,

        # The application identifier URI.
        [Parameter(Mandatory=$true, ParameterSetName='ByUri')]
        [ValidateNotNullOrEmpty()]
        [string] $ApplicationIdentifierUri,

        # The identifier of the object (user, service principal, etc.) to which ownership of the target application should be granted.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $OwnerObjectId
    )

    # Lookup the target objects
    $params      = if ($ApplicationId) {@{ AppId = $ApplicationId }} else {@{ AppUri = $ApplicationIdentifierUri }}
    $application = Get-GraphApplication @params -ErrorAction Stop
    $owner       = Invoke-GraphApi -ApiPath "directoryObjects/$OwnerObjectId" -ErrorAction Stop

    # Lookup the existing owners and grant ownership if not already granted
    $owners = (Invoke-GraphApi -Method Get -ApiPath "applications/$($application.objectId)/owners" -ErrorAction Stop).value
    if ($owners | Where objectId -EQ $OwnerObjectId)
    {
        Write-Verbose "Object '$($owner.objectId)' of type '$($owner.objectType)' is already an owner of the application '$($application.displayName)' ($($application.appId))" -Verbose
    }
    else
    {
        Write-Verbose "Granting ownership of application '$($application.displayName)' ($($application.appId)) to object '$($owner.objectId)' of type '$($owner.objectType)'." -Verbose
        Invoke-GraphApi -Method Post -ApiPath "applications/$($application.objectId)/`$links/owners" -Verbose -ErrorAction Stop -Body (ConvertTo-Json ([pscustomobject]@{
            url = '{0}/directoryObjects/{1}' -f $Script:GraphEnvironment.GraphEndpoint.AbsoluteUri.TrimEnd('/'), $OwnerObjectId
        }))
    }
}

<#
.Synopsis
   Updates the set of client certificates (key credentials) usable by an application in Graph.
#>
function Set-GraphApplicationClientCertificates
{
    [CmdletBinding(DefaultParameterSetName='ById')]
    param
    (
        # The application identifier.
        [Parameter(Mandatory=$true, ParameterSetName='ById')]
        [ValidateNotNullOrEmpty()]
        [string] $ApplicationId,

        # The application identifier URI.
        [Parameter(Mandatory=$true, ParameterSetName='ByUri')]
        [ValidateNotNullOrEmpty()]
        [string] $ApplicationIdentifierUri,

        # The client certificates used to authenticate with graph as the application / service principal.
        [Parameter(Mandatory=$true)]
        [ValidateNotNull()]
        [System.Security.Cryptography.X509Certificates.X509Certificate2[]] $ClientCertificate = @()
    )

    # Lookup the target application
    $params      = if ($ApplicationId) {@{ AppId = $ApplicationId }} else {@{ AppUri = $ApplicationIdentifierUri }}
    $application = Get-GraphApplication @params -ErrorAction Stop

    # Initialize the application key credentials with which it can authenticate
    $requestBody = @{
        'keyCredentials@odata.type' = 'Collection(Microsoft.DirectoryServices.KeyCredential)'
        'keyCredentials'            = @()
    }

    foreach ($cert in $ClientCertificate)
    {
        $customKeyIdentifier = [Convert]::ToBase64String($cert.GetCertHash())
        if (-not (@($requestBody['keyCredentials']) | Where customKeyIdentifier -EQ $customKeyIdentifier))
        {
            Write-Verbose "Adding key credential for application using client certificate '$($cert.Subject)' ($($cert.Thumbprint))" -Verbose

            $requestBody['keyCredentials'] += @(,([pscustomobject]@{
                keyId               = [Guid]::NewGuid()
                type                = "AsymmetricX509Cert"
                usage               = "Verify"
                customKeyIdentifier = $customKeyIdentifier
                value               = [Convert]::ToBase64String($cert.GetRawCertData())
                startDate           = $cert.NotBefore.ToUniversalTime().ToString('o')
                endDate             = $cert.NotAfter.ToUniversalTime().ToString('o')
            }))
        }
        else
        {
            Write-Verbose "Key credentials already added to application for client certificate '$($cert.Subject)' ($($cert.Thumbprint))" -Verbose
        }
    }

    $requestBodyAsJson = $requestBody | ConvertTo-Json -Depth 10
    $apiPath           = "directoryObjects/$($application.objectId)/Microsoft.DirectoryServices.Application"
    
    Write-Verbose "Updating key credentials on application '$($application.displayName)' ($($application.appId))..." -Verbose
    $noResponse = Invoke-GraphApi -Method Patch -ApiPath $apiPath -Body $requestBodyAsJson -Verbose -ErrorAction Stop
}

<#
.Synopsis
   Creates a new self-signed Json Web Token to use as a client assertion or in other Graph API calls.
#>
function New-SelfSignedJsonWebToken
{
    [CmdletBinding()]
    param
    (
        # The client certificate used to sign the token.
        [Parameter(Mandatory=$true)]
        [ValidateNotNull()]
        [System.Security.Cryptography.X509Certificates.X509Certificate2] $ClientCertificate,

        # The client ID (appId) for the token.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $ClientId,

        # The target audience for the token.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $Audience,

        # The number of seconds relative to the current UTC datetime before which the token will be invalid. Default is -90 (90 seconds ago from 'now').
        [Parameter()]
        [int] $NotBeforeSecondsRelativeToNow = -90,

        # The number of seconds relative to the current UTC datetime until which the token will be valid. Default is 3600 (one hour from 'now').
        [Parameter()]
        [int] $ExpirationSecondsRelativeToNow = 3600
    )

    function ConvertTo-Base64UrlEncode([byte[]]$bytes) { [System.Convert]::ToBase64String($bytes).Replace('/','_').Replace('+','-').Trim('=') }

    $tokenHeaders = [ordered]@{
        alg = 'RS256'
        x5t = ConvertTo-Base64UrlEncode $ClientCertificate.GetCertHash()
    }

    $currentUtcDateTimeInSeconds = ([datetime]::UtcNow - [datetime]'1970-01-01 00:00:00').TotalSeconds

    $tokenClaims = [ordered]@{
        aud = $Audience
        exp = [long]($currentUtcDateTimeInSeconds + $ExpirationSecondsRelativeToNow)
        iss = $ClientId
        jti = [guid]::NewGuid().ToString()
        nbf = [long]($currentUtcDateTimeInSeconds + $NotBeforeSecondsRelativeToNow)
        sub = $ClientId
    }

    Write-Verbose "Preparing client assertion with token header: '$(ConvertTo-Json $tokenHeaders -Compress)' and claims: $(ConvertTo-Json $tokenClaims)"

    # Note - we escape the forward slashes ('/') as the ConvertTo-Json cmdlet does not. This may not actually be necessary.
    $tokenParts = @()
    $tokenParts += ConvertTo-Base64UrlEncode ([System.Text.Encoding]::UTF8.GetBytes((ConvertTo-Json $tokenHeaders -Depth 10 -Compress).Replace('/', '\/')))
    $tokenParts += ConvertTo-Base64UrlEncode ([System.Text.Encoding]::UTF8.GetBytes((ConvertTo-Json $tokenClaims -Depth 10 -Compress).Replace('/', '\/')))

    $sha256Hash = ''
    $sha256 = [System.Security.Cryptography.SHA256]::Create()
    try
    {
        $sha256Hash = $sha256.ComputeHash([System.Text.Encoding]::UTF8.GetBytes($tokenParts -join '.'))
    }
    finally
    {
        if ($sha256) { $sha256.Dispose(); $sha256 = $null }
    }

    # Note - the default instance of the RSACryptoServiceProvider instantiated on the client certificate may only support SHA1.
    # E.g. Even when "$($ClientCertificate.SignatureAlgorithm.FriendlyName)" evaluates to "sha256RSA", the value of
    # "$($ClientCertificate.PrivateKey.SignatureAlgorithm)" may evaulate to "http://www.w3.org/2000/09/xmldsig#rsa-sha1".
    # Furthermore, the private key is likely not marked as exportable, so we cannot "simply" instantiate a new RSACryptoServiceProvider instance.
    # We must first create new CSP parameters with a "better" cryptographic service provider that supports SHA256, and use those parameters
    # to instantiate a "better" RSACryptoServiceProvider which also supports SAH256. Failure to do this will result in the following error:
    # "Exception calling "CreateSignature" with "1" argument(s): "Invalid algorithm specified."
    # It may be possible to bypass this issue of the certificate is generated with the "correct" cryptographic service provider, but if the certificate
    # was created by a CA or if the provider type was not the "correct" type, then this workaround must be used.
    # Note - this assumes certificate is installed in the local machine store.
    $csp = New-Object System.Security.Cryptography.CspParameters(
        ($providerType=24),
        ($providerName='Microsoft Enhanced RSA and AES Cryptographic Provider'),
        $ClientCertificate.PrivateKey.CspKeyContainerInfo.KeyContainerName)
    $csp.Flags = [System.Security.Cryptography.CspProviderFlags]::UseMachineKeyStore

    $signatureBytes = $null
    $rsa = New-Object System.Security.Cryptography.RSACryptoServiceProvider($csp)
    try
    {
        $signatureBytes = $rsa.SignHash($sha256Hash, [System.Security.Cryptography.HashAlgorithmName]::SHA256, [System.Security.Cryptography.RSASignaturePadding]::Pkcs1)
    }
    finally
    {
        if ($rsa) { $rsa.Dispose(); $rsa = $null }
    }

    $tokenParts += ConvertTo-Base64UrlEncode $signatureBytes

    return ($tokenParts -join '.')
}

<#
.Synopsis
   Adds a new client certificate to a graph application / service principal if it is not already added.
#>
function Add-GraphApplicationClientCertificate
{
    [CmdletBinding()]
    [OutputType([string])]
    param
    (
        # The application identifier.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $ApplicationId,

        # A client certificate used to authenticate with graph as the application / service principal.
        [Parameter(Mandatory=$true)]
        [ValidateNotNull()]
        [System.Security.Cryptography.X509Certificates.X509Certificate2] $CurrentClientCertificate,

        # The new client certificate to add to be used to authenticate with graph as the application / service principal.
        [Parameter(Mandatory=$true)]
        [ValidateNotNull()]
        [System.Security.Cryptography.X509Certificates.X509Certificate2] $NewClientCertificate
    )

    # https://msdn.microsoft.com/en-us/Library/Azure/Ad/Graph/api/functions-and-actions#AddKey

    $application         = Invoke-GraphApi -ApiPath "applicationsByAppId/$ApplicationId" -ErrorAction Stop
    $customKeyIdentifier = [Convert]::ToBase64String($NewClientCertificate.GetCertHash())

    if (($keyCredential = $application.keyCredentials | Where customKeyIdentifier -EQ $customKeyIdentifier))
    {
        Write-Verbose "Application '$($application.displayName)' ($ApplicationId) already has certificate '$($NewClientCertificate.Thumbprint)' added under keyId '$($keyCredential.keyId)' and customKeyIdentifier '$customKeyIdentifier'" -Verbose
        Write-Verbose "keyCredential: $(ConvertTo-Json $keyCredential -Depth 4)"
        return $keyCredential.keyId
    }

    $params = @{
        ClientCertificate = $CurrentClientCertificate
        ClientId          = $ApplicationId

        # Audience needs to be AAD Graph SPN
        Audience = (Get-GraphEnvironmentInfo).Applications.WindowsAzureActiveDirectory.Id

        # The token lifespan should not exceed 10 minutes. Where token lifespan is the difference between EXP and NBF claims.
        NotBeforeSecondsRelativeToNow  = -90
        ExpirationSecondsRelativeToNow = 500
    }
    $jwt = New-SelfSignedJsonWebToken @params

    $params = @{
        Method  = [Microsoft.PowerShell.Commands.WebRequestMethod]::Post
        ApiPath = "applicationsByAppId/$ApplicationId/addKey"
        Body    = (ConvertTo-Json -Depth 4 -Compress ([pscustomobject]@{
            keyCredential = @{
                type                = "AsymmetricX509Cert"
                usage               = "Verify"
                customKeyIdentifier = $customKeyIdentifier
                value               = [Convert]::ToBase64String($NewClientCertificate.GetRawCertData())
                startDate           = $NewClientCertificate.NotBefore.ToUniversalTime().ToString('o')
                endDate             = $NewClientCertificate.NotAfter.ToUniversalTime().ToString('o')
            }
            proof = "Bearer $jwt"
        }))
    }

    $response = Invoke-GraphApi @params -ErrorAction Stop
    $keyId    = $response.value[0].keyId
    Write-Verbose "Response: $(ConvertTo-Json $response -Depth 4)"
    Write-Verbose "Client certificate added to application '$($application.displayName)' ($ApplicationId) with thumbprint '$($NewClientCertificate.Thumbprint)' under keyId '$keyId' and customKeyIdentifier '$customKeyIdentifier'" -Verbose
    return $keyId
}

<#
.Synopsis
   Removes a client certificate from a graph application / service principal based on the associated keyId.
#>
function Remove-GraphApplicationClientCertificate
{
    [CmdletBinding()]
    param
    (
        # The application identifier.
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [string] $ApplicationId,

        # A client certificate used to authenticate with graph as the application / service principal.
        [Parameter(Mandatory=$true)]
        [ValidateNotNull()]
        [System.Security.Cryptography.X509Certificates.X509Certificate2] $CurrentClientCertificate,

        # The client certificate to remove from the application to no-longer be used to authenticate with graph as the application / service principal.
        [Parameter(Mandatory=$true)]
        [ValidateNotNull()]
        [System.Security.Cryptography.X509Certificates.X509Certificate2] $ClientCertificateToRemove
    )

    # https://msdn.microsoft.com/en-us/Library/Azure/Ad/Graph/api/functions-and-actions#removeKey

    $application         = Invoke-GraphApi -ApiPath "applicationsByAppId/$ApplicationId" -ErrorAction Stop
    $customKeyIdentifier = [Convert]::ToBase64String($ClientCertificateToRemove.GetCertHash())

    if (-not ($keyCredential = $application.keyCredentials | Where customKeyIdentifier -EQ $customKeyIdentifier))
    {
        Write-Verbose "Application '$($application.displayName)' ($ApplicationId) does not have certificate '$($ClientCertificateToRemove.Thumbprint)' added under customKeyIdentifier '$customKeyIdentifier' or has already had this certificate ." -Verbose
        return
    }

    $params = @{
        ClientCertificate = $CurrentClientCertificate
        ClientId          = $ApplicationId

        # Audience needs to be AAD Graph SPN
        Audience = (Get-GraphEnvironmentInfo).Applications.WindowsAzureActiveDirectory.Id

        # The token lifespan should not exceed 10 minutes. Where token lifespan is the difference between EXP and NBF claims.
        NotBeforeSecondsRelativeToNow  = -90
        ExpirationSecondsRelativeToNow = 500
    }
    $jwt = New-SelfSignedJsonWebToken @params

    $params = @{
        Method  = [Microsoft.PowerShell.Commands.WebRequestMethod]::Post
        ApiPath = "applicationsByAppId/$ApplicationId/removeKey"
        Body    = (ConvertTo-Json -Depth 4 -Compress ([pscustomobject]@{
            keyId = $keyCredential.keyId
            proof = "Bearer $jwt"
        }))
    }

    $noResponse = Invoke-GraphApi @params -ErrorAction Stop
    Write-Verbose "Removed client certificate on application '$($application.displayName)' ($ApplicationId) with thumbprint '$($ClientCertificateToRemove.Thumbprint)' under keyId '$($keyCredential.keyId)' and customKeyIdentifier '$customKeyIdentifier'" -Verbose
}

[System.Reflection.Assembly]::LoadWithPartialName('System.Web') | Out-Null

<#
.Synopsis
   Formats the provided query string parameters into a URL-encoded query string format.
#>
function ConvertTo-QueryString
{
    [CmdletBinding()]
    param
    (
        [Parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]
        [HashTable] $QueryParameters = @{}
    )

    $query = [System.Web.HttpUtility]::ParseQueryString("?")
    $QueryParameters.GetEnumerator() | ForEach { $query.Add($_.Key, $_.Value) }
    Write-Output $query.ToString()
}

Export-ModuleMember -Function @(
    'Initialize-GraphEnvironment'
    'Get-GraphEnvironmentInfo'
    #'Assert-GraphEnvironmentIsInitialized'
    #'Assert-GraphConnection'
    'Get-GraphToken'
    'Update-GraphAccessToken'
    'Invoke-GraphApi'
    'Get-GraphDefaultVerifiedDomain'
    'Find-GraphApplication'
    'Get-GraphApplication'
    'Remove-GraphObject'
    'Find-GraphApplicationDataByServicePrincipalTag'
    #'Get-GraphObjectWithRetry'
    'Get-GraphApplicationServicePrincipal'
    'Initialize-GraphApplicationServicePrincipal'
    'Update-GraphApplicationServicePrincipalTags'
    'Initialize-GraphOAuth2PermissionGrant'
    'Initialize-GraphAppRoleAssignment'
    'Initialize-GraphDirectoryRoleMembership'
    'New-GraphPermissionDescription'
    'Get-GraphApplicationPermissions'
    'Grant-GraphApplicationPermission'
    'Grant-GraphApplicationPermissions'
    'Show-GraphApplicationPermissionDescriptions'
    'Initialize-GraphApplication'
    'Initialize-GraphApplicationOwner'
    'Set-GraphApplicationClientCertificates'
    'New-SelfSignedJsonWebToken'
    'Add-GraphApplicationClientCertificate'
    'Remove-GraphApplicationClientCertificate'
    #'ConvertTo-QueryString'
)

# SIG # Begin signature block
# MIInnwYJKoZIhvcNAQcCoIInkDCCJ4wCAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCC+5erMpcXpQ/A8
# YPqswYOUZfIJd5+IgDuJ0XjTKBerQ6CCDYEwggX/MIID56ADAgECAhMzAAACzI61
# lqa90clOAAAAAALMMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMjIwNTEyMjA0NjAxWhcNMjMwNTExMjA0NjAxWjB0MQsw
# CQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9u
# ZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMR4wHAYDVQQDExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIB
# AQCiTbHs68bADvNud97NzcdP0zh0mRr4VpDv68KobjQFybVAuVgiINf9aG2zQtWK
# No6+2X2Ix65KGcBXuZyEi0oBUAAGnIe5O5q/Y0Ij0WwDyMWaVad2Te4r1Eic3HWH
# UfiiNjF0ETHKg3qa7DCyUqwsR9q5SaXuHlYCwM+m59Nl3jKnYnKLLfzhl13wImV9
# DF8N76ANkRyK6BYoc9I6hHF2MCTQYWbQ4fXgzKhgzj4zeabWgfu+ZJCiFLkogvc0
# RVb0x3DtyxMbl/3e45Eu+sn/x6EVwbJZVvtQYcmdGF1yAYht+JnNmWwAxL8MgHMz
# xEcoY1Q1JtstiY3+u3ulGMvhAgMBAAGjggF+MIIBejAfBgNVHSUEGDAWBgorBgEE
# AYI3TAgBBggrBgEFBQcDAzAdBgNVHQ4EFgQUiLhHjTKWzIqVIp+sM2rOHH11rfQw
# UAYDVR0RBEkwR6RFMEMxKTAnBgNVBAsTIE1pY3Jvc29mdCBPcGVyYXRpb25zIFB1
# ZXJ0byBSaWNvMRYwFAYDVQQFEw0yMzAwMTIrNDcwNTI5MB8GA1UdIwQYMBaAFEhu
# ZOVQBdOCqhc3NyK1bajKdQKVMFQGA1UdHwRNMEswSaBHoEWGQ2h0dHA6Ly93d3cu
# bWljcm9zb2Z0LmNvbS9wa2lvcHMvY3JsL01pY0NvZFNpZ1BDQTIwMTFfMjAxMS0w
# Ny0wOC5jcmwwYQYIKwYBBQUHAQEEVTBTMFEGCCsGAQUFBzAChkVodHRwOi8vd3d3
# Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NlcnRzL01pY0NvZFNpZ1BDQTIwMTFfMjAx
# MS0wNy0wOC5jcnQwDAYDVR0TAQH/BAIwADANBgkqhkiG9w0BAQsFAAOCAgEAeA8D
# sOAHS53MTIHYu8bbXrO6yQtRD6JfyMWeXaLu3Nc8PDnFc1efYq/F3MGx/aiwNbcs
# J2MU7BKNWTP5JQVBA2GNIeR3mScXqnOsv1XqXPvZeISDVWLaBQzceItdIwgo6B13
# vxlkkSYMvB0Dr3Yw7/W9U4Wk5K/RDOnIGvmKqKi3AwyxlV1mpefy729FKaWT7edB
# d3I4+hldMY8sdfDPjWRtJzjMjXZs41OUOwtHccPazjjC7KndzvZHx/0VWL8n0NT/
# 404vftnXKifMZkS4p2sB3oK+6kCcsyWsgS/3eYGw1Fe4MOnin1RhgrW1rHPODJTG
# AUOmW4wc3Q6KKr2zve7sMDZe9tfylonPwhk971rX8qGw6LkrGFv31IJeJSe/aUbG
# dUDPkbrABbVvPElgoj5eP3REqx5jdfkQw7tOdWkhn0jDUh2uQen9Atj3RkJyHuR0
# GUsJVMWFJdkIO/gFwzoOGlHNsmxvpANV86/1qgb1oZXdrURpzJp53MsDaBY/pxOc
# J0Cvg6uWs3kQWgKk5aBzvsX95BzdItHTpVMtVPW4q41XEvbFmUP1n6oL5rdNdrTM
# j/HXMRk1KCksax1Vxo3qv+13cCsZAaQNaIAvt5LvkshZkDZIP//0Hnq7NnWeYR3z
# 4oFiw9N2n3bb9baQWuWPswG0Dq9YT9kb+Cs4qIIwggd6MIIFYqADAgECAgphDpDS
# AAAAAAADMA0GCSqGSIb3DQEBCwUAMIGIMQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMTIwMAYDVQQDEylNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0
# ZSBBdXRob3JpdHkgMjAxMTAeFw0xMTA3MDgyMDU5MDlaFw0yNjA3MDgyMTA5MDla
# MH4xCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdS
# ZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMT
# H01pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBIDIwMTEwggIiMA0GCSqGSIb3DQEB
# AQUAA4ICDwAwggIKAoICAQCr8PpyEBwurdhuqoIQTTS68rZYIZ9CGypr6VpQqrgG
# OBoESbp/wwwe3TdrxhLYC/A4wpkGsMg51QEUMULTiQ15ZId+lGAkbK+eSZzpaF7S
# 35tTsgosw6/ZqSuuegmv15ZZymAaBelmdugyUiYSL+erCFDPs0S3XdjELgN1q2jz
# y23zOlyhFvRGuuA4ZKxuZDV4pqBjDy3TQJP4494HDdVceaVJKecNvqATd76UPe/7
# 4ytaEB9NViiienLgEjq3SV7Y7e1DkYPZe7J7hhvZPrGMXeiJT4Qa8qEvWeSQOy2u
# M1jFtz7+MtOzAz2xsq+SOH7SnYAs9U5WkSE1JcM5bmR/U7qcD60ZI4TL9LoDho33
# X/DQUr+MlIe8wCF0JV8YKLbMJyg4JZg5SjbPfLGSrhwjp6lm7GEfauEoSZ1fiOIl
# XdMhSz5SxLVXPyQD8NF6Wy/VI+NwXQ9RRnez+ADhvKwCgl/bwBWzvRvUVUvnOaEP
# 6SNJvBi4RHxF5MHDcnrgcuck379GmcXvwhxX24ON7E1JMKerjt/sW5+v/N2wZuLB
# l4F77dbtS+dJKacTKKanfWeA5opieF+yL4TXV5xcv3coKPHtbcMojyyPQDdPweGF
# RInECUzF1KVDL3SV9274eCBYLBNdYJWaPk8zhNqwiBfenk70lrC8RqBsmNLg1oiM
# CwIDAQABo4IB7TCCAekwEAYJKwYBBAGCNxUBBAMCAQAwHQYDVR0OBBYEFEhuZOVQ
# BdOCqhc3NyK1bajKdQKVMBkGCSsGAQQBgjcUAgQMHgoAUwB1AGIAQwBBMAsGA1Ud
# DwQEAwIBhjAPBgNVHRMBAf8EBTADAQH/MB8GA1UdIwQYMBaAFHItOgIxkEO5FAVO
# 4eqnxzHRI4k0MFoGA1UdHwRTMFEwT6BNoEuGSWh0dHA6Ly9jcmwubWljcm9zb2Z0
# LmNvbS9wa2kvY3JsL3Byb2R1Y3RzL01pY1Jvb0NlckF1dDIwMTFfMjAxMV8wM18y
# Mi5jcmwwXgYIKwYBBQUHAQEEUjBQME4GCCsGAQUFBzAChkJodHRwOi8vd3d3Lm1p
# Y3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY1Jvb0NlckF1dDIwMTFfMjAxMV8wM18y
# Mi5jcnQwgZ8GA1UdIASBlzCBlDCBkQYJKwYBBAGCNy4DMIGDMD8GCCsGAQUFBwIB
# FjNodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2RvY3MvcHJpbWFyeWNw
# cy5odG0wQAYIKwYBBQUHAgIwNB4yIB0ATABlAGcAYQBsAF8AcABvAGwAaQBjAHkA
# XwBzAHQAYQB0AGUAbQBlAG4AdAAuIB0wDQYJKoZIhvcNAQELBQADggIBAGfyhqWY
# 4FR5Gi7T2HRnIpsLlhHhY5KZQpZ90nkMkMFlXy4sPvjDctFtg/6+P+gKyju/R6mj
# 82nbY78iNaWXXWWEkH2LRlBV2AySfNIaSxzzPEKLUtCw/WvjPgcuKZvmPRul1LUd
# d5Q54ulkyUQ9eHoj8xN9ppB0g430yyYCRirCihC7pKkFDJvtaPpoLpWgKj8qa1hJ
# Yx8JaW5amJbkg/TAj/NGK978O9C9Ne9uJa7lryft0N3zDq+ZKJeYTQ49C/IIidYf
# wzIY4vDFLc5bnrRJOQrGCsLGra7lstnbFYhRRVg4MnEnGn+x9Cf43iw6IGmYslmJ
# aG5vp7d0w0AFBqYBKig+gj8TTWYLwLNN9eGPfxxvFX1Fp3blQCplo8NdUmKGwx1j
# NpeG39rz+PIWoZon4c2ll9DuXWNB41sHnIc+BncG0QaxdR8UvmFhtfDcxhsEvt9B
# xw4o7t5lL+yX9qFcltgA1qFGvVnzl6UJS0gQmYAf0AApxbGbpT9Fdx41xtKiop96
# eiL6SJUfq/tHI4D1nvi/a7dLl+LrdXga7Oo3mXkYS//WsyNodeav+vyL6wuA6mk7
# r/ww7QRMjt/fdW1jkT3RnVZOT7+AVyKheBEyIXrvQQqxP/uozKRdwaGIm1dxVk5I
# RcBCyZt2WwqASGv9eZ/BvW1taslScxMNelDNMYIZdDCCGXACAQEwgZUwfjELMAkG
# A1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQx
# HjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEoMCYGA1UEAxMfTWljcm9z
# b2Z0IENvZGUgU2lnbmluZyBQQ0EgMjAxMQITMwAAAsyOtZamvdHJTgAAAAACzDAN
# BglghkgBZQMEAgEFAKCBrjAZBgkqhkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAcBgor
# BgEEAYI3AgELMQ4wDAYKKwYBBAGCNwIBFTAvBgkqhkiG9w0BCQQxIgQgfQ1mxRwy
# vzRPHZ1PPrMjwL2lpxpdUPFPMcsfCpJoNDgwQgYKKwYBBAGCNwIBDDE0MDKgFIAS
# AE0AaQBjAHIAbwBzAG8AZgB0oRqAGGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbTAN
# BgkqhkiG9w0BAQEFAASCAQCCvmYLCRCt4F21T8U1WnUeH5FgXnKjCHEvXxjT49oK
# cbuf1jTvnI8lgCzwVjQI8kksbYBg7UkW4f61aXmkhVGF/nU2bVXJtbHlY8gOZuWX
# d/6B9l6eZAxRtUPfQJJ7lOjL15kGPRHRxzhq78we/dJVWACiGrMpzgDcNM4WbgTk
# FqLn2nr/ygM0EaYxSC8hAypbRpdbWPseOerxZGNa0ZytqBKc4diSg7oIAjk6xa/8
# +Jdn7xGBdmPpbSS6hqrWdcZBLODvpS8630j6As1/7drTwoxyHAdDyeVjrWIlGtx2
# 7lvzGddlz0bpYVsR1kmuFSpAe5poKXezAoTqQQcq28bOoYIW/jCCFvoGCisGAQQB
# gjcDAwExghbqMIIW5gYJKoZIhvcNAQcCoIIW1zCCFtMCAQMxDzANBglghkgBZQME
# AgEFADCCAVAGCyqGSIb3DQEJEAEEoIIBPwSCATswggE3AgEBBgorBgEEAYRZCgMB
# MDEwDQYJYIZIAWUDBAIBBQAEIAIkrfTXdXvEntADRz802mBJPjxXF+MQwMHsM640
# hX4/AgZjIzXQL/kYEjIwMjIwOTE1MjIxNDA2LjA3WjAEgAIB9KCB0KSBzTCByjEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjElMCMGA1UECxMcTWlj
# cm9zb2Z0IEFtZXJpY2EgT3BlcmF0aW9uczEmMCQGA1UECxMdVGhhbGVzIFRTUyBF
# U046MTJCQy1FM0FFLTc0RUIxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1w
# IFNlcnZpY2WgghFWMIIHDDCCBPSgAwIBAgITMwAAAaEBhVWZuVRdigABAAABoTAN
# BgkqhkiG9w0BAQsFADB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3Rv
# bjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0
# aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMDAeFw0y
# MTEyMDIxOTA1MjRaFw0yMzAyMjgxOTA1MjRaMIHKMQswCQYDVQQGEwJVUzETMBEG
# A1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWlj
# cm9zb2Z0IENvcnBvcmF0aW9uMSUwIwYDVQQLExxNaWNyb3NvZnQgQW1lcmljYSBP
# cGVyYXRpb25zMSYwJAYDVQQLEx1UaGFsZXMgVFNTIEVTTjoxMkJDLUUzQUUtNzRF
# QjElMCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZTCCAiIwDQYJ
# KoZIhvcNAQEBBQADggIPADCCAgoCggIBANrJPF7la6SStjHFW4cthb0ERYIP2SSO
# ecew4rAZ10g9tmUtj6Xmi8sM1/4EQxAoBlAjcNf3WXXIIO4/1fu048LjxlEZcD9t
# /2qXQUrnjfyAiXtxXnbxd4Q4XBz8D5RshR9bb3o6aDxnrbFpC/eOsbhT+muICsX9
# 6vVQDUc24gZTKqOXKCJI/ArY2cUCmLUkP5R5/lzjuSHulbUqPtGdyGkV5j0x6Q9B
# GJrtwRpRhTiyoKIlV0Mml58u89P0R22GVDHvmV3H4DBl/Zr1Pu5BFIGHy2nE90gM
# OQqJYzCMpOsBjT0Dcj+OJ2o+5zw+9f6yrGrJkQ3aHgYDQR2OaTrieQi6QArXwrmc
# AsMs71IxPGkDBAgdEO1l5MKW8A8ISjLW+08Pt/56oepK2675cKR9GNcSlf36H1+u
# wHT8GAPkIF/cQssBrxN58x8dlYQlFM82ttcwqLNKtRKRW//cc/9mwmnBrPkzLZFv
# JzcCH1tPvp4EmTJ9PkU32/8pDQefGFEyzoceFOY3H4vO1hyL68d/QPdAfV4KNlZl
# GOnWY7LGk9TaYMwbqB6W8mx7UnNEAOjtgiiT8ncJxubwxsFubzmKiAWW0Ud5wcUQ
# XCuwMYEWc1gcyFxtqtA0D6BjZ7aX18CRfcyMjtSSWSjPvj8/ooip7mNx30U8JttJ
# tgf04uy155g5AgMBAAGjggE2MIIBMjAdBgNVHQ4EFgQUFr8gMttjjvlVDIqJlLDj
# uXT9zKkwHwYDVR0jBBgwFoAUn6cVXQBeYl2D9OXSZacbUzUZ6XIwXwYDVR0fBFgw
# VjBUoFKgUIZOaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9jcmwvTWlj
# cm9zb2Z0JTIwVGltZS1TdGFtcCUyMFBDQSUyMDIwMTAoMSkuY3JsMGwGCCsGAQUF
# BwEBBGAwXjBcBggrBgEFBQcwAoZQaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3Br
# aW9wcy9jZXJ0cy9NaWNyb3NvZnQlMjBUaW1lLVN0YW1wJTIwUENBJTIwMjAxMCgx
# KS5jcnQwDAYDVR0TAQH/BAIwADATBgNVHSUEDDAKBggrBgEFBQcDCDANBgkqhkiG
# 9w0BAQsFAAOCAgEAIMMjESV9eUblgpwqss9JL5/clZmmvAoSIBK+K9odMFGVe0Cz
# 5ORp1ywR6L73Dm+YXm0JTNMfhzScIjttnGewv5LpeyI6zdnVXhZf4jChdQnjMu+z
# T6ZPi+MYO1h8pD9uyYkpqvZz32b98e/VabYJNzJp4++LzomHdTIuN1EgtZu3Ozig
# iYUzDApvMd0+inGsGGCL4LVhmyGixYuWDPK7GNSX6o2DWbnYwmZ/XWWgjsP0cmhp
# DN36t/3bxjyu9QuaDaH8bnSj4PRQnUVr9wklod8Hex8rD1foau1dgaOYzf6D4CFp
# Wx+6kpc204W7m2csq8Afk4iMQNhXVgqaVe4G6FthqyzKA8UyY2AbYCeTd2sRwNxm
# EJdeqlGzM2jUXoa7kkKlBlds4kz1R7k+Ukq2YiYBggazD6mcfL+vmCBJg6niDlnW
# hT0aFUIzdXRP1p157o5RcGTWsTh1lz9Sw+WPSqiKWMv6U3UDmCSabPuTm0g5tUYH
# t0l3PwnQXBdETmpi7UB29q5VtnAZCQvXHxor+y+MRBbQ1TInb3OcMeJeXm8uhFOO
# MWmyFQGLb4hj6Y2psuaPbiPl5P5uMOUTceY20s+ktwwNipnUf7pTpiZqI2ZpzaNF
# cMBp4QT+6gMy3Z0Ct8k/sz4wO/fPM01Mg1640S0NWCb3PB+AhQJecNm5W2Iwggdx
# MIIFWaADAgECAhMzAAAAFcXna54Cm0mZAAAAAAAVMA0GCSqGSIb3DQEBCwUAMIGI
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMTIwMAYDVQQDEylN
# aWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0ZSBBdXRob3JpdHkgMjAxMDAeFw0yMTA5
# MzAxODIyMjVaFw0zMDA5MzAxODMyMjVaMHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQI
# EwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3Nv
# ZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBD
# QSAyMDEwMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA5OGmTOe0ciEL
# eaLL1yR5vQ7VgtP97pwHB9KpbE51yMo1V/YBf2xK4OK9uT4XYDP/XE/HZveVU3Fa
# 4n5KWv64NmeFRiMMtY0Tz3cywBAY6GB9alKDRLemjkZrBxTzxXb1hlDcwUTIcVxR
# MTegCjhuje3XD9gmU3w5YQJ6xKr9cmmvHaus9ja+NSZk2pg7uhp7M62AW36MEByd
# Uv626GIl3GoPz130/o5Tz9bshVZN7928jaTjkY+yOSxRnOlwaQ3KNi1wjjHINSi9
# 47SHJMPgyY9+tVSP3PoFVZhtaDuaRr3tpK56KTesy+uDRedGbsoy1cCGMFxPLOJi
# ss254o2I5JasAUq7vnGpF1tnYN74kpEeHT39IM9zfUGaRnXNxF803RKJ1v2lIH1+
# /NmeRd+2ci/bfV+AutuqfjbsNkz2K26oElHovwUDo9Fzpk03dJQcNIIP8BDyt0cY
# 7afomXw/TNuvXsLz1dhzPUNOwTM5TI4CvEJoLhDqhFFG4tG9ahhaYQFzymeiXtco
# dgLiMxhy16cg8ML6EgrXY28MyTZki1ugpoMhXV8wdJGUlNi5UPkLiWHzNgY1GIRH
# 29wb0f2y1BzFa/ZcUlFdEtsluq9QBXpsxREdcu+N+VLEhReTwDwV2xo3xwgVGD94
# q0W29R6HXtqPnhZyacaue7e3PmriLq0CAwEAAaOCAd0wggHZMBIGCSsGAQQBgjcV
# AQQFAgMBAAEwIwYJKwYBBAGCNxUCBBYEFCqnUv5kxJq+gpE8RjUpzxD/LwTuMB0G
# A1UdDgQWBBSfpxVdAF5iXYP05dJlpxtTNRnpcjBcBgNVHSAEVTBTMFEGDCsGAQQB
# gjdMg30BATBBMD8GCCsGAQUFBwIBFjNodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20v
# cGtpb3BzL0RvY3MvUmVwb3NpdG9yeS5odG0wEwYDVR0lBAwwCgYIKwYBBQUHAwgw
# GQYJKwYBBAGCNxQCBAweCgBTAHUAYgBDAEEwCwYDVR0PBAQDAgGGMA8GA1UdEwEB
# /wQFMAMBAf8wHwYDVR0jBBgwFoAU1fZWy4/oolxiaNE9lJBb186aGMQwVgYDVR0f
# BE8wTTBLoEmgR4ZFaHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJv
# ZHVjdHMvTWljUm9vQ2VyQXV0XzIwMTAtMDYtMjMuY3JsMFoGCCsGAQUFBwEBBE4w
# TDBKBggrBgEFBQcwAoY+aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0
# cy9NaWNSb29DZXJBdXRfMjAxMC0wNi0yMy5jcnQwDQYJKoZIhvcNAQELBQADggIB
# AJ1VffwqreEsH2cBMSRb4Z5yS/ypb+pcFLY+TkdkeLEGk5c9MTO1OdfCcTY/2mRs
# fNB1OW27DzHkwo/7bNGhlBgi7ulmZzpTTd2YurYeeNg2LpypglYAA7AFvonoaeC6
# Ce5732pvvinLbtg/SHUB2RjebYIM9W0jVOR4U3UkV7ndn/OOPcbzaN9l9qRWqveV
# tihVJ9AkvUCgvxm2EhIRXT0n4ECWOKz3+SmJw7wXsFSFQrP8DJ6LGYnn8AtqgcKB
# GUIZUnWKNsIdw2FzLixre24/LAl4FOmRsqlb30mjdAy87JGA0j3mSj5mO0+7hvoy
# GtmW9I/2kQH2zsZ0/fZMcm8Qq3UwxTSwethQ/gpY3UA8x1RtnWN0SCyxTkctwRQE
# cb9k+SS+c23Kjgm9swFXSVRk2XPXfx5bRAGOWhmRaw2fpCjcZxkoJLo4S5pu+yFU
# a2pFEUep8beuyOiJXk+d0tBMdrVXVAmxaQFEfnyhYWxz/gq77EFmPWn9y8FBSX5+
# k77L+DvktxW/tM4+pTFRhLy/AsGConsXHRWJjXD+57XQKBqJC4822rpM+Zv/Cuk0
# +CQ1ZyvgDbjmjJnW4SLq8CdCPSWU5nR0W2rRnj7tfqAxM328y+l7vzhwRNGQ8cir
# Ooo6CGJ/2XBjU02N7oJtpQUQwXEGahC0HVUzWLOhcGbyoYICzTCCAjYCAQEwgfih
# gdCkgc0wgcoxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJTAj
# BgNVBAsTHE1pY3Jvc29mdCBBbWVyaWNhIE9wZXJhdGlvbnMxJjAkBgNVBAsTHVRo
# YWxlcyBUU1MgRVNOOjEyQkMtRTNBRS03NEVCMSUwIwYDVQQDExxNaWNyb3NvZnQg
# VGltZS1TdGFtcCBTZXJ2aWNloiMKAQEwBwYFKw4DAhoDFQAbcXaM7gsQxUvCAoZd
# 1gw3gUGA4KCBgzCBgKR+MHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5n
# dG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9y
# YXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQSAyMDEwMA0G
# CSqGSIb3DQEBBQUAAgUA5s20STAiGA8yMDIyMDkxNTIyMjUxM1oYDzIwMjIwOTE2
# MjIyNTEzWjB2MDwGCisGAQQBhFkKBAExLjAsMAoCBQDmzbRJAgEAMAkCAQACAQgC
# Af8wBwIBAAICEaQwCgIFAObPBckCAQAwNgYKKwYBBAGEWQoEAjEoMCYwDAYKKwYB
# BAGEWQoDAqAKMAgCAQACAwehIKEKMAgCAQACAwGGoDANBgkqhkiG9w0BAQUFAAOB
# gQCNsuV74Qop7NgQcbmo4LG2vKcPxDo0tb/Ba9Paehl2Wexhyg54D6b0pcrxxycr
# xHxAzKXIa/EoIeRXz+z5Ys2yqWQX0jp4IUWc8lAZiTlxcyvRnhd+QytxB5O62aSZ
# NM+l18Zv6nSDlP76KQ5kZuO6Q6wlYRMMRyiyKIPtueFiqjGCBA0wggQJAgEBMIGT
# MHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdS
# ZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMT
# HU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQSAyMDEwAhMzAAABoQGFVZm5VF2KAAEA
# AAGhMA0GCWCGSAFlAwQCAQUAoIIBSjAaBgkqhkiG9w0BCQMxDQYLKoZIhvcNAQkQ
# AQQwLwYJKoZIhvcNAQkEMSIEIOeJQu6aaj0JybOk6Luw9cyNONMiULQfbsRdvo2u
# y7Q8MIH6BgsqhkiG9w0BCRACLzGB6jCB5zCB5DCBvQQg6whU8TqBgmgggo6EcgXt
# SUkKzCXggk8hK84oid+O0IQwgZgwgYCkfjB8MQswCQYDVQQGEwJVUzETMBEGA1UE
# CBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9z
# b2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGltZS1TdGFtcCBQ
# Q0EgMjAxMAITMwAAAaEBhVWZuVRdigABAAABoTAiBCB/fCZQQqPu4VLdpQFxifBz
# vASKhPTB8RVxtrPOcJTLtDANBgkqhkiG9w0BAQsFAASCAgB1JHyQDgFqGnMLY5eB
# 2S0Lxixb2Qe5bJQiB+xBXyaM9O2SEQIqYZv2N8d28CcL30kv43DvhKNMEqFNwQqY
# tgWa0P7+eFL9nr/jyFjDEbYtjsKEe68mq6hn8E7TniOJ6VgFdhwnZUAXgK1P3oXh
# wi9zfc/354v+8PfEiD4Not/YJry36HkzEBLzYRBP+XSJhIhAwKDKacJaVyJkKBaS
# 5fubD8L3Z4qf7hUqTrMJH0Xd1kiSN/u2y/Yik61r2N7t3n0jYZSDe8qCMPpTWpNa
# 6ztJf1l1B/+BWY2rzEIhgflR2Wu/RFnF03GXtanZ2c8Feke5d69TV3xQ7uyoFLxd
# CT6y3hg39LRBWZVElJ3FhKyLxjz/r0z3MtkNbUZKz8PKY8gonnLSvP1RBEkFktfl
# 08SKhCOxaZgaDg8nlTETukWtK0b9+hGR4LPbPUBhi//yVjOsj5RnJre+yC6SPU6Z
# 0U2N9wdrvUflNE1Yy4+4KmXXT2dcGO/eoAUkaDaEw3Od14Zx5QhRSfa3r634rJkr
# iEQHJ5/Pd4XhWSyu3bAP3KNR+UmjowLWI24qjjnVhYD/zLS1OJjfHPwHXzJgmonX
# f1uUzL60TIxxP8kZrJQXHzr0JjpPRMbwHviX4U0bL+1tysDs6+RlqpkN5pTjrK7L
# UuVcJBA67VntTjZWSf4haz+TbQ==
# SIG # End signature block
